#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <values.h>
#include <assert.h>

#include <libgnomeprint/gnome-print.h>
#include <glib.h>
#include <slang.h>

#include <bond/dbobject.h>
#include <bond/db.h>

#include "db.h"
#include "layout.h"
#include "print.h"

#define min(a,b) ((a)<(b))?(a):(b)
#define max(a,b) ((a)>(b))?(a):(b)

typedef enum
   {
   LT_image,
   LT_text,
   LT_date,
   LT_money,
   LT_paragraph,
   LT_padding,
   LT_shape /* drawing objects, ie: line/box etc...
	     * To use it, define "data" as something like:
	     *   "data=line,100", which will give you a 100 long line
	     * wherever that object is supposed to be
	     */
   } layout_type;

typedef enum
   {
   LS_line,
   LS_box
   } layout_shape;

typedef enum
   {
   LP_leftof,
   LP_rightof,
   LP_above,
   LP_below
   } layout_position;

typedef enum
   {
   LA_top,
   LA_bottom,
   LA_vcentered
   } layout_valignment;

typedef enum
   {
   LA_left,
   LA_right,
   LA_hcentered
   } layout_halignment;

typedef struct
   {
   double red;
   double green;
   double blue;
   } layout_colour;

typedef struct layout layout;
typedef struct layout_object layout_object;

struct layout_object
   {
   layout *l;                                       /* pointer to base list */
   char *name;                /* name, used to identify the field from bond */
   char *data;              /* data to use if this is not a database object *
                             * basically, if this is not NULL, data will be *
			     * taken from here, rather than the bond object */
   layout_type type;                       /* type of object to expect here */
   layout_object *parent;                                  /* parent object */
   layout_position pos;                        /* relative to parent object */
   layout_valignment valign; 
   layout_halignment halign;
   int has_colour;
   layout_colour colour;                               /* foreground colour */
   int border;                          /* border width (0 means no border) */
   int has_background;                          /* do we have a background? */
   layout_colour background;           /* background colour (if applicable) */
   int min_width;    /* min width field can have. Used for packing & tables *
 -1 => auto calculated, all obj of this type have same width for this field */
   int min_height;
   char *fontname;
   int font_size;
   GnomeFont *font;
   int underline;
   
   int width; /* Values for layout. Calculated in layout_compile */
   int height; 
   int xpos;
   int ypos;   
   };

struct layout
   {
   int total_width;
   int total_height;
   GList *objects;
   };

static layout_type decode_type (char *type)
  {
  if (strcasecmp (type, "image") == 0)
     return LT_image;
  else if (strcasecmp (type, "text") == 0 || strcasecmp (type, "string") == 0)
     return LT_text;
  else if (strcasecmp (type, "date") == 0)
    return LT_date;
  else if (strcasecmp (type, "money") == 0)
    return LT_money;
  else if (strcasecmp (type, "paragraph") == 0)
    return LT_paragraph;   
  else if (strcasecmp (type, "padding") == 0)
    return LT_padding;
  else if (strcasecmp (type, "shape") == 0)
    return LT_shape;
  
  return LT_text;
}

/* decodes a string of the form "a=b" into its two parts, "a", and "b".
 * These values are stored in the arrays passed in
 */
static void decode_string (char *string, char *part1, char *part2)
   {
   char *equals;
   int pos;
   
   equals = strstr (string, "=");
   
   if (!equals)
      {
      part1[0] = '\0';
      part2[0] = '\0';
      return;
      }
   
   pos = equals - string;
   
   strncpy (part1, string, pos);
   part1[pos] = '\0';
   strcpy (part2, &string[pos + 1]);
   
/*   fprintf (stderr, "decode_string: decoded %s into '%s' = '%s'\n",
	    string, part1, part2); */
   }

layout_valignment decode_valign (char *str)
   {
   if (strcasecmp (str, "top") == 0)
      return LA_top;
   else if (strcasecmp (str, "bottom") == 0)
      return LA_bottom;
   else if (strcasecmp (str, "centered") == 0)
      return LA_vcentered;
   return LA_vcentered;
   }

layout_halignment decode_halign (char *str)
   {
   if (strcasecmp (str, "left") == 0)
      return LA_left;
   else if (strcasecmp (str, "right") == 0)
      return LA_right;
   else if (strcasecmp (str, "centered") == 0)
      return LA_hcentered;
   return LA_hcentered;
   }

void decode_colour (char *string, layout_colour *colour)
   {
   colour->red = colour->green = colour->blue = 0.0;
   
   colour->red = atof (string);
   string = strstr (string, ",");
   if (!string)
      return;
   
   colour->green = atof (&string[1]);
   string = strstr (&string[1], ",");
   if (!string)
      return;
   
   colour->blue = atof (&string[1]);
   }

/* builds a layout object from a slang call. Expects there to be some 
 * information in the SLang stack, for calls to SLpop_string
 */ 
static layout_object *layout_object_build (char *name, char *type, 
					   char **args, int num,
					   layout_object *parent)
   {
   layout_object *retval;
   char *string;
   char part1[80], part2[80];
   int i;
   
   retval = malloc (sizeof (layout_object));
   
   retval->type = decode_type (type);
   retval->name = strdup (name);

   retval->data = NULL;
   
   retval->valign = LA_vcentered;
   retval->halign = LA_hcentered;
   retval->has_colour = 0;
   decode_colour ("0,0,0", &retval->colour);
   retval->border = 0;
   retval->has_background = 0;
   decode_colour ("0,0,0", &retval->background);
   retval->min_width = -1;
   retval->min_height = -1;
   if (parent)
      {
      retval->fontname = parent->fontname;
      retval->font_size = parent->font_size;
      }
   else
      {
      retval->fontname = "Nimbus Sans L Bold";
      retval->font_size = 10;
      }
   retval->underline = 0;
   
   for (i = 0; i < num; i++)
      {
      string = args[i];

      /*fprintf (stderr, "popped string: %s\n", string);*/
      
      decode_string (string, part1, part2);
      
      if (strcmp (part1, "data") == 0)
	 retval->data = strdup (part2);
      else if (strcmp (part1, "valign") == 0)
	 retval->valign = decode_valign (part2);
      else if (strcmp (part1, "halign") == 0)
	 retval->halign = decode_halign (part2);
      else if (strcmp (part1, "colour") == 0)
	 {
	 decode_colour (part2, &retval->colour);
	 retval->has_colour = 1;
	 }
      else if (strcmp (part1, "border") == 0)
	 retval->border = atoi (part2);
      else if (strcmp (part1, "backgroundcolour") == 0)
	 {
	 decode_colour (part2, &retval->background);
	 retval->has_background = 1;
	 }
      else if (strcmp (part1, "font") == 0)
	 retval->fontname = strdup (part2);
      else if (strcmp (part1, "size") == 0)
	 retval->font_size = atoi (part2);
      else if (strcmp (part1, "underline") == 0)
	 retval->underline =  (strcasecmp (part2, "on") == 0);
      else if (strcmp (part1, "minwidth") == 0)
	{
	  if (part2[strlen(part2) - 1] == '%')
	    retval->min_width = (int) ((double) print_paper_width () * (atof (part2) / 100.0));
	  else
	    retval->min_width = atoi (part2);
	}
      else if (strcmp (part1, "minheight") == 0)
	{
	  if (part2[strlen (part2) - 1] == '%')
	    retval->min_width = (int)((double) print_paper_height () * (atof (part2) / 100.0));
	  else
	    retval->min_height = atoi (part2);
	}
      else
	 fprintf (stderr, "Unknown layout object: '%s' = '%s'\n",
		  part1, part2);      
      }

   retval->font = gnome_font_new (retval->fontname, retval->font_size);
   
   return retval;
   }

int layout_begin (void)
   {
   char *name, *type;
   layout *l;
   layout_object *o;
   int num, i;
   char **args;

   num = SLang_Num_Function_Args;     
   
   if (num < 2)
      {
      fprintf (stderr, "layout->begin requires 2 arguments, name and type\n");
      return -1;
      }
   
   args = malloc (sizeof (char *) * num - 2);
   
   /* Build array of var args, ignoring the 2 required, name & type */
   for (i = 0; i < num - 2; i++) 
      if (SLpop_string (&args[i]) == -1)
	 {
	 fprintf (stderr, "layout->begin Arg: %d invalid\n", i + 2);
	 args[i] = NULL;
	 }

   if (SLpop_string (&type) == -1)
      fprintf (stderr, "layout->begin: type must be a string\n");
   
   if (SLpop_string (&name) == -1)
      fprintf (stderr, "layout->begin: name must be a string\n");
   
   l = malloc (sizeof (layout));
   o = layout_object_build (name, type, args, num - 2, NULL);
   o->parent = NULL;
   
    /* doesn't matter, since the first object is not positioned 
     * relative to anything */
   o->pos = LP_leftof;
   
   for (i = 0; i < num - 2; i++)
      free (args [i]);
   free (args);   
   
   if (o)
      {
      l->objects = g_list_append (NULL, o);
      o->l = l;
      return (int)o;
      }
   fprintf (stderr, "Error building object\n");
   return 0;
   }

int layout_add (void)
   {
   int relative, num, i;
   char *name, *type, *position, **args;
   layout_object *o, *next;
   layout *l;
   
/*   SLdo_pop_n (4); */
   
   num = SLang_Num_Function_Args;     
   
   if (num < 4)
      {
      fprintf (stderr, "layout->add requires 4 arguments: relative, name, type and position\n");
      return -1;
      }
   
   args = malloc (sizeof (char *) * num - 4);
   
   /* Build array of var args, ignoring the 2 required, name & type */
   for (i = 0; i < num - 4; i++) 
      if (SLpop_string (&args[i]) == -1)
	 {
	 fprintf (stderr, "layout->begin Arg: %d invalid\n", i + 4);
	 args[i] = NULL;
	 }

   if (SLpop_string (&position) == -1)
      fprintf (stderr, "layout->add: position must be a string\n");
   
   if (SLpop_string (&type) == -1)
      fprintf (stderr, "layout->add: type must be a string\n");
   
   if (SLpop_string (&name) == -1)
      fprintf (stderr, "layout->add: name must be a string\n");
   
   if (SLang_pop_integer (&relative) == -1)
      fprintf (stderr, "layout->add: relative must be an integer\n");
   
   o = (void *)relative;
   l = o->l;
   
   next = layout_object_build (name, type, args, num - 4, o);
   if (next)
      {
      next->l = l;
      next->parent = o;
      l->objects = g_list_append (l->objects, next);
      
      if (strcmp (position, "leftof") == 0)
	 next->pos = LP_leftof;
      else if (strcmp (position, "rightof") == 0)
	 next->pos = LP_rightof;
      else if (strcmp (position, "above") == 0)
	 next->pos = LP_above;
      else if (strcmp (position, "below") == 0)
	 next->pos = LP_below;
      else
	 {
	 fprintf (stderr, "Invalid position identifier: %s\n", position);
	 next->pos = LP_leftof;
	 }
	 
      return (int)next;
      }
   fprintf (stderr, "Error building object\n");   
   return 0;
   }

 /* returns the relavent string for a given object. Maybe data, or maybe
  * database info */
static char *get_object_string (layout_object *o, Object *obj)
   {
   char *retval;
   
   if (o->data)
      retval =o->data;
   else
      db_getfieldstr (obj, o->name, &retval);
   
   return retval;
   } 

static layout_shape get_shape_type (layout_object *o)
   {
   assert (o->type == LT_shape);   
   assert (o->data);
   
   if (strncasecmp (o->data, "line,", 5) == 0)
      return LS_line;
   else if (strncasecmp (o->data, "box,", 4) == 0)
      return LS_box;
   return LS_line;
   }

static void set_width_height (layout_object *o, Object *obj)
   {
   switch (o->type)
      {      
      case LT_text :
	 {
	 char *str = get_object_string (o, obj);	 
	 o->width = print_linewidth ((void *)&o->font, str);
	 if (o->min_width != -1)
	    o->width = max (o->width, o->min_width);
	 o->height = print_lineheight ((void *)&o->font);
	 break;
	 }

      case LT_padding :
	{
	  if (o->min_width == -1 && o->min_height == -1)
	    fprintf (stderr, "Padding element specified with no minimum width or height\n");
	  o->height = max (0, o->min_height);
	  o->width = max (0, o->min_height);
	  break;
	}
      
      case LT_shape:
	 {
	 if (o->parent)
	    o->width = o->parent->width;
	 else
	    o->width = 100;
	 o->height = 0;
	 break;
	 }
      
      default :
	 {
	 fprintf (stderr, "set_width_height: object type unsupported: %d\n",
		  o->type);
	 o->width = o->height = 0;
	 }
      }
   }

static void set_x_y (layout_object *o)
   {
   if (!o->parent)
      {
      o->xpos = 0;
      o->ypos = 0;
      return;
      }

   switch (o->pos)
      {
      case LP_leftof:
	 {
	 o->xpos = o->parent->xpos - o->width;
	 o->ypos = o->parent->ypos;
	 break;
	 }
      case LP_rightof:      
	 {
	 o->xpos = o->parent->xpos + o->parent->width;
	 o->ypos = o->parent->ypos;
	 break;
	 }
      case LP_above:
	 {
	 o->xpos = o->parent->xpos;
	 o->ypos = o->parent->ypos - o->height;
	 break;
	 }
      case LP_below:
	 {
	 o->xpos = o->parent->xpos;
	 o->ypos = o->parent->ypos + o->parent->height;
	 break;
	 }
      }
   }

static int get_min_x (layout *l)
   {
   GList *list;
   layout_object *o;
   int minx = MAXINT;
   
   for (list = g_list_first (l->objects); list != NULL; list = list->next)
      {
      o = list->data;
      minx = min(minx, o->xpos);
      }
   
   return minx;
   }

static int get_min_y (layout *l)
   {
   GList *list;
   layout_object *o;
   int miny = MAXINT;
   
   for (list = g_list_first (l->objects); list != NULL; list = list->next)
      {
      o = list->data;      
      miny = min(miny, o->ypos);
      }
   
   return miny;
   }

static void normalize_coords (layout *l)
   {
   GList *list;
   layout_object *o;
   int minx, miny;

   minx = get_min_x (l);
   miny = get_min_y (l);
   
   for (list = g_list_first (l->objects); list != NULL; list = list->next)
      {
      o = list->data;
      
      o->xpos -= minx;
      o->ypos -= miny;
      }
   }

static void set_width_height_total (layout *l)
   {
   layout_object *o;
   GList *list;
   int max_width = MININT, max_height = MININT;
   
   for (list = g_list_first (l->objects); list != NULL; list = list->next)
      {
      o = list->data;
      
      max_width = max(o->xpos + o->width, max_width);
      max_height = max(o->ypos + o->height, max_height);
      }
   
   l->total_width = max_width;
   l->total_height = max_height;
   }

/* pretty inefficient, since it traverses the list about 4 times, it only
 * really has to do it twice, once to compile & find minimum x&y, then once
 * again to normalize it... never mind, fix it if we discover it is slow
 */

void layout_compile (int *lay, int *dbo)
   {
   layout_object *o = (void *)*lay;
   layout *l = o->l;
   GList *cur;
   Object *obj = (void *)*dbo;
   
   for (cur = g_list_first (l->objects); cur != NULL; cur = cur->next)
      {
      o = cur->data;

      set_width_height (o, obj);
      
      set_x_y (o);      
      }
   
   normalize_coords (l);
   
   set_width_height_total (l);
   }

/* Calculates the size of the layout, given some data (from compile) */
int layout_width (int *lay)
   {
   layout_object *o = (void *)*lay;
   return o->l->total_width;
   }

int layout_height (int *lay)
   {
   layout_object *o = (void *)*lay;
   return o->l->total_height;
   }

void object_display (int *printer, layout_object *o, Object *obj, int x, int y)
   {
   if (o->has_colour)
      print_setcolour (printer, &o->colour.red, &o->colour.green, 
		       &o->colour.blue);
   
   switch (o->type)
      {
      case LT_text:
	 {
	 int xpos, ypos;
	 
	 xpos = x + o->xpos;
	 ypos = y - o->ypos;
	 print_showat (printer, (void *)&o->font, &xpos, &ypos,
		       get_object_string (o, obj));
	 break;
	 }
	 
      case LT_padding: break;
      
      case LT_shape:
	 {
	 layout_shape type = get_shape_type (o);
	 
	 switch (type)
	    {
	    case LS_line :
	       {
	       int tx, ty;
	       
	       tx = x + o->width;
	       ty = y - o->height;
	       print_newpath (printer);
	       print_moveto (printer, &x, &y);
	       print_lineto (printer, &tx, &ty);
	       print_stroke (printer);	       
	       break;
	       }
	    
	    case LS_box :
	       {
	       break;
	       }
	    }
	 
	 break;
	 }
      
      default :
	 {
	 fprintf (stderr, "object_display: type not supported: %d\n",
		  o->type);
	 }
      }
   }

/* draws the layout onto the printer */
void layout_display (int *printer, int *lay, int *dbo, int *x, int *y)
   {
   layout_object *o = (void *)*lay, *cur;
   layout *l = o->l;
   Object *obj = (void *)*dbo;
   GList *next;
   
   for (next = g_list_first (l->objects); next != NULL; next = next->next)
      {
      cur = next->data;
      object_display (printer, cur, obj, *x, *y);
      }
      
   }

#define pi(a) if (SLang_pop_integer (& a) == -1) \
                 fprintf (stderr, "Invalid arguments to full display\n")
#define ps(a) if (SLpop_string (& a) == -1) \
                 fprintf (stderr, "Invalid arguments to full display\n")


void layout_fulldisplay (void)
   {
   int i, num, pos;
   char *headcall, *footcall;
   int printer, dbo, lay, x, y, bottom;
   char *header, *footer;

   if (SLang_Num_Function_Args != 8)
      {
      fprintf (stderr, "fulldisplay takes 8 arguments\n");
      return;
      }
      
   
   ps(footer);
   ps(header);
   pi(bottom);
   pi(y);
   pi(x);
   pi(lay);
   pi(dbo);
   pi(printer);
   
   headcall = NULL;
   if (header)
      {
      if (header[0] != '\0')
	 asprintf (&headcall, "%s();", header);
      free (header);
      }

   footcall = NULL;
   if (footer)
      {
      if (footer[0] != '\0')
	 asprintf (&footcall, "%s();", footer);
      free (footer);
      }
   
   num = db_objnum_wrap (&dbo);
   
   pos = y;
   
   if (headcall)
      SLang_load_string (headcall);
   
   for (i = 0; i < num; i++)
      {
      db_moverecord_wrap (&dbo, &i);
      layout_compile (&lay, &dbo);
      
      if (pos - layout_height (&lay) < bottom)
	 {
	 if (footcall)
	    SLang_load_string (footcall);
	 print_showpage (&printer);
	 print_newpage (&printer);
	 if (headcall)
	    SLang_load_string (headcall);
	 pos = y;
	 }

      layout_display (&printer, &lay, &dbo, &x, &pos);
      pos -= layout_height (&lay);
      }

   if (footcall)
      SLang_load_string (footcall);
   
   
   if (headcall)
      free (headcall);
   if (footcall)
      free (footcall);
   
   }
