/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <glib.h>

#include "bond.h"
#include "db.h"
#include "ui.h"
#include "bc.h"
#include "dropdown.h"
#include "bondxml.h"
#include "bondrelation.h"
#include "biglists.h"
#include "postit.h"
#include "watchtower.h"
#include "fieldpath.h"
#include "bondtrigger.h"

/**
 * bond_test:
 * @testnum: Test number to execute.
 *
 * Run a series of internal tests.
 *
 * Returns: non-zero on error.
 */
gint
bond_test(Bond * bond, gint testnum)
	{
	gint retval = 0, passed = 0;
	BondObject *bobj, *tmpbobj;
	BondRelation *brel;
	UiWidget *widget;

	widget = ui_getwidget(bond->uixml, "mainwindow");

	switch (testnum)
		{
	case 0:
		retval = bondxml_test(bond, 0);
		break;
	case 1:
		test_annonce(1, "BondObject creation and distruction");
		brel = bondrelation_create(RelationLoadAll);
		bond_spawn(bond, NULL, "mainwindow", brel, FALSE);
		bobj = bond_getbyname(bond, "mainwindow/vbox0/cell20");
		if (bobj != NULL)
			passed++;
		else
			errormsg("Unable to create bobj");
		if (bobj->brel != NULL)
			passed++;
		else
			errormsg("Failed to generate database object, no "
			         "bondrelation");
		if (bobj->obj != NULL)
			{
			passed++;
			if (bobj->obj->num == 4)
				passed++;
			else
				errormsg("Bondobject donest contain 4 rows.");
			}
		else
			errormsg("Failed to generate database object, its NULL");
		if (bobj->children != NULL)
			passed++;
		else
			errormsg("No children on this widget??");
		retval = test_total(passed, 5);
		break;
	case 2:
		test_annonce(2, "BondObject loading");
		brel = bondrelation_create(RelationLoadAll);
		bond_spawn(bond, NULL, "mainwindow", brel, FALSE);
		tmpbobj = bond_getbyname(bond, "mainwindow/vbox0/box11/id");
		test_result(tmpbobj->obj->basetable, "soldier", &passed);
		test_result(tmpbobj->oprnode->field, "id", &passed);
		if (tmpbobj->fieldactive == TRUE)
			passed++;
		tmpbobj = bond_getbyname(bond, "mainwindow/vbox0/box11/name");
		g_assert(tmpbobj);
		g_assert(tmpbobj->oprnode);
		test_result(tmpbobj->oprnode->field, "name", &passed);
		tmpbobj = bond_getbyname(bond, "mainwindow/vbox0/box11/dob");
		g_assert(tmpbobj);
		test_result(tmpbobj->oprnode->field, "dob", &passed);
		tmpbobj = bond_getbyname(bond, "mainwindow/vbox0/box11/kills");
		g_assert(tmpbobj);
		test_result(tmpbobj->oprnode->field, "kills", &passed);
		tmpbobj = bond_getbyname(bond, "mainwindow");
		bond_close(tmpbobj, TRUE);
		retval = test_total(passed, 6);
		break;
	case 3:
		test_annonce(3, "BondObject form showing");
		brel = bondrelation_create(RelationLoadAll);
		bobj = bond_spawn(bond, NULL, "mainwindow", brel, FALSE);

		bond_paint(bobj, FALSE, TRUE);
		debuglog(30, "Now displaying searchwindow.");
		bobj = bond_spawn(bond, NULL, "searchwindow", NULL, FALSE);
		bond_paint(bobj, FALSE, TRUE);
		ui_main(bond->uixml);
		break;
	case 4:
		test_annonce(4, "BondObject loadby across form.");
		bondxml_test(bond, 1);
		bobj = bond_spawn(bond, NULL, "mainwindow", NULL, FALSE);
		bobj = bond_getbyname(bond, "mainwindow/vbox0/cell31");
		if (!bobj)
			errormsg("Error finding mainwindow/vbox0/cell31");
		else if (bobj->obj != NULL && bobj->obj->num == 2)
			passed++;
		else
			warningmsg("Loadby failed, only returned %d results",
			           bobj->obj->num);
		retval = test_total(passed, 1);
		break;
	case 5:
		fieldpath_test(0);
		retval = test_total(passed, 0);
		break;
	default:
		errormsg("Unknown test number");
		retval = 1;
		break;
		}
	return retval;
	}

static void
bond_debug_rec(BondObject * bobj, gint indent)
	{
	GList *walk;
	gchar *tmpstr;

	if (bobj == NULL)
		{
		debug_output("%*sNULL bobj. Ouch.\n", indent, "");
		return ;
		}
	debug_output("%*s%s\n", indent, "", bobj->widgetname);
	if (bobj->fieldactive == TRUE)
		{
		tmpstr = NULL;
		if (bobj->obj != NULL)
			tmpstr = bobj->obj->basetable;
		debug_output("%*sField: %s from %s\n", indent, "",
		             bobj->oprnode->field, tmpstr);
		}
	if (bobj->dbobjactive == TRUE)
		{
		tmpstr = NULL;
		if (bobj->obj != NULL)
			tmpstr = bobj->obj->basetable;
		debug_output("%*sdbobject: %s from %s\n", indent, "",
		             bobj->oprnode->dbobj->name, tmpstr);
		}
	for (walk = g_list_first(bobj->children); walk != NULL; walk = walk->next)
		bond_debug_rec(walk->data, indent + 1);
	}

/**
 * bond_debug:
 * @bobj: Bond object to show debugging output for.
 * 
 * Will show debugging information about the @bobj. All variables in the bond 
 * object is printed and it recursively goes down the bond object tree.
 * 
 * Returns: Nothing
 */
void
bond_debug(BondObject * bobj)
	{
	debug_minorheader();
	bond_debug_rec(bobj, 0);
	debug_minorfooter();
	}

/* ============================================================
 * Code starts here...  Initializing and cleaning.
 * ============================================================
 */

static void
bond_cleanuphash_item(void *ptr)
	{
	g_list_free((GList *) ptr);
	}

/**
 * bond_inithash:
 * @bond: The global bond structure.
 *
 * Initialise the hash tables.  Two hash tables, globalbobj and 
 * globalbobj_byobj is created.  These two hash tables are used to store
 * all bond objects and for lookups.
 * 
 * Returns: 0 if two hash tables created successfully. -1 if either
 * hash tables in the bond structure already exist.
 */
gint
bond_inithash(Bond * bond)
	{
	gint retval = 0;

	g_assert(bond);

	/* If either global bond object hash tables already exist, return -1 */
	if (bond->globalbobj || bond->globalbobj_byobj)
		{
		warningmsg("bond->globalbobj is already initialised");
		return -1;
		}

	/* Else create two new global bond object hash tables and return 0 */
	bond->globalbobj = g_hash_table_new_full(g_str_hash, g_str_equal, NULL,
	                   bond_free);
	bond->globalbobj_byobj = g_hash_table_new_full(NULL, NULL, NULL,
	                         bond_cleanuphash_item);

	return retval;
	}

/**
 * bond_cleanuphash:
 * @bond: Global bond object to do the cleanesing on.
 * 
 * All the bond objects registered will be freed, and closed.  This is normally
 * called on programme exit.
 *
 * Returns: Nothing at all.
 */
void
bond_cleanuphash(Bond * bond)
	{
	if (bond->globalbobj)
		g_hash_table_destroy(bond->globalbobj);
	if (bond->globalbobj_byobj)
		g_hash_table_destroy(bond->globalbobj_byobj);
	if (bond->objcache)
		g_hash_table_destroy(bond->objcache);
	}

/**
 * bond_create:
 * @bond: Global bond object.
 * @oprnode: Operations node.
 * 
 * Given an OperationsNode @oprnode, bond_create will create a new bond object
 * based on that oprnode.  The oprnode is generated directly from the bond xml
 * document. Once the bond object is created it can be painted and accessed.
 * This function is called internally by bond_spawn().
 * 
 * Returns: Newly created bond object to be freed by bond_free().
 */
BondObject *
bond_create(Bond * bond, OperationsNode * oprnode)
	{
	BondObject *bobj;

	g_assert(bond);
	g_assert(oprnode);

	bobj = mem_calloc(sizeof(BondObject));
	bobj->oprnode = oprnode;
	bobj->widget = oprnode->widget;
	bobj->widgetname = oprnode->widgetname;
	bobj->widgettype = oprnode->widgettype;
	bobj->bond = bond;

	g_hash_table_insert(bond->globalbobj, bobj->widgetname, bobj);
	return bobj;
	}

/**
 * bond_getbyname:
 * @widgetname: Name of a widget specified in xml file to retrieve a bond
 * object from.
 *
 * Retrieves a bond object if @widgetname points to a widget name,
 * otherwise fails and returns %NULL.  It looks it up by the longname first, else
 * it will revert to the widget shortname.  It will store the last item retrieved
 * as a cache item.
 *
 * Returns: BondObject or a %NULL if not found.
 */
BondObject *
bond_getbyname(Bond * bond, gchar * widgetname)
	{
	BondObject *bobj;
	gchar *longwidgetname;

	g_assert(bond->globalbobj);
	g_assert(widgetname);
	if (widgetname[0] == 0)
		{
		errormsg("No widgetname specified.");
		return NULL;
		}
	/* Check cached value */
	if (bond->lastbobj && strcmp(widgetname, bond->lastlongwidgetname) == 0)
		return bond->lastbobj;

	if ((bobj = g_hash_table_lookup(bond->globalbobj, widgetname)) == NULL)
		{
		longwidgetname = bondxml_getwidgetlongname(bond, widgetname);
		if (longwidgetname == NULL)
			{
			errormsg("Failed to find widget %s.", widgetname);
			return NULL;
			}
		bobj = g_hash_table_lookup(bond->globalbobj, longwidgetname);
		}

	/* set temporary cache */
	if (bobj)
		{
		bond->lastbobj = bobj;
		bond->lastlongwidgetname = widgetname;
		}
	return bobj;
	}

/**
 * bond_getbydbobjectsrc:
 * @bond:
 * @dbobjectsrc:
 *
 * Looks for any dbobject's with a name of @dbobjectsrc
 * and returns the BondObject they are contained within.
 */
BondObject *
bond_getbydbobjectsrc(Bond * bond, gchar * dbobjectsrc)
	{
	BondObject *retbobj = NULL;

	g_assert(bond->globalbobj);
	g_assert(dbobjectsrc);
	if (dbobjectsrc[0] == 0)
		{
		errormsg("No dbobjectsrc specified.");
		return NULL;
		}
	if ((retbobj = g_hash_table_lookup(bond->globalbobj_byobj, dbobjectsrc))
	        == NULL)
		{
		debugmsg("Failed to find dbobjectsrc %s", dbobjectsrc);
		}
	return retbobj;
	}

/**
 * bond_isfieldactive:
 * @bobj: Bond Object.
 * 
 * Field active is if a widget is associated with a database field.  If there 
 * are no valid database record currently loaded isfieldactive will return 
 * false. It is only set to true if there is the possiblity of saving database 
 * back to the database if a <field> tag is specified in the xml defination.
 * 
 * Returns: True or False.
 */
gboolean
bond_isfieldactive(BondObject * bobj)
	{
	if (bobj->fieldactive == TRUE || bobj->active == TRUE)
		return TRUE;
	return FALSE;
	}

/**
 * bond_isparentobj:
 * @bobj: Bond Object.
 * 
 * Checks if the @bobj has a database object attached to it (a dbobject
 * tag in the bond xml document).  A database object is used to create
 * a dbobj from bonddb and could often be a sql statement.
 * 
 * Returns: True or False.
 */
gboolean
bond_isparentobj(BondObject * bobj)
	{
	g_assert(bobj);
	if (bobj->oprnode->dbobj != NULL)
		return TRUE;
	return FALSE;
	}

/* close any subforms that may have been opened by this form */
static gint
bond_freeassocate_objects(BondObject * bobj)
	{
	gboolean done = FALSE;
	GList *walk, *bobjlist;
	BondObject *tmpbobj;

	if (!bobj->obj || !bobj->oprnode->dbobj)
		return 1;

	while (done == FALSE)
		{
		/* normally you do one iternation of the link list */
		done = TRUE;
		bobjlist =
		    g_hash_table_lookup(bobj->bond->globalbobj_byobj, bobj->obj);
		/* m.essage("freeing associated objects for %s. %d items",
		   bobj->widgetname,g_list_length(bobjlist)); */
		for (walk = g_list_first(bobjlist); walk != NULL; walk = walk->next)
			{
			tmpbobj = walk->data;
			if (!tmpbobj)
				errormsg("tmpbobj is null?");
			if (tmpbobj->parent == NULL)
				{
				debuglog(20, "Going to close %s", 
						tmpbobj->widgetname);
				bond_close(tmpbobj, TRUE);
				/* This may seem weird, but basically what i
				 * am doing is start at the top off of 
				 * the link list again */
				done = FALSE;
				break;
				}
			}
		}
	return 0;
	}

/**
 * bond_free:
 * @ptr: Bond object, not of type void * but actually of type BondObject*.
 * 
 * Free's a bondobject, this is automatically called when the bond object is 
 * removed from the hash table.  Dont call this function directly, call 
 * g_hash_table_remove() instead.  All children of this object are also
 * closed.  This will default to not saving the contents of any database
 * objects assocated with the bond object @ptr.
 * 
 * Returns: Nothing.
 */
void
bond_free(void *ptr)
	{
	GList *walk, *children;
	BondObject *bobj = (BondObject *) ptr;

	/* make sure pointer is valid */
	mem_verify(bobj);
	/* m.essage("Freeing %s",bobj->widgetname); */
	/* Free any signals that may be attached to this if it has a widget */
	postit_sig_disconnect(bobj);
	/* Unregister the bond object, watchtowers and registerors. */
	bond_unregisterobj(bobj);
	registration_remove_bybobj(bobj);
	watchtower_unregister(bobj);
	/* Clear the last item of cache incase it interfers */
	if (bobj->bond->lastlongwidgetname &&
	        strcmp(bobj->bond->lastlongwidgetname, bobj->widgetname) == 0)
		{
		bobj->bond->lastbobj = NULL;
		bobj->bond->lastlongwidgetname = NULL;
		}
	/* If dropdown box remove hash tables associated with dropdowns */
	dropdown_free(bobj);
	/* If have a bond relation, free that */
	if (bobj->brel != NULL)
		{
		if (bondrelation_free(bobj->brel) == 0)
			{
			/* Free database object if exist and it isn't perment 
			 * cache. */
			if (bobj->obj && bobj->oprnode->dbobj->cache == FALSE)
				{
				/* m.essage("freeing dbobj instance
				   %d",bobj->obj->objectinstance); */
				db_freeobject(bobj->obj);
				}
			}
		}
	/* Deal with parents */
	children = g_list_first(bobj->children);
	bobj->children = NULL;
	/* Remove us from our parents list of children. We are forgotten */
	if (bobj->parent && bobj->parent->children)
		g_list_remove(bobj->parent->children, bobj);
	/* Free all the children. This is all widgets enclosed in this widget */
	for (walk = children; walk != NULL; walk = walk->next)
		bond_close(walk->data, FALSE);
	/* Close any objects that relate to this. */
	bond_freeassocate_objects(bobj);
	/* Clear g_lists */
	if (bobj->functions)
		g_list_free(bobj->functions);
	if (bobj->cousins)
		g_list_free(bobj->cousins);
	if (bobj->dependon)
		g_list_free(bobj->dependon);
	if (bobj->dependents)
		g_list_free(bobj->dependents);
	if (bobj->registrationlist)
		g_list_free(bobj->registrationlist);
	if (bobj->watchtowerlist)
		g_list_free(bobj->watchtowerlist);
	/* children are set to null early on so you know to disown from 
	 * parent */
	if (children)
		g_list_free(children);
	/* Clean out the garbage */
	if (bobj->garbage)
		mem_freegarbage(bobj->garbage);
	/* Finally free the bond object */
	mem_free(bobj);
	}

/* ============================================================
 * Bond object showing and hiding.
 * ============================================================
 */

/**
 * bond_show:
 * @bobj: Bond object.
 * @recusiveinprogress: If all objects which are children of @bobj are 
 * also to be made visible.
 * 
 * Display a bondobject making it visible. bond_paint() actually displays
 * the data in the form, bond_show simply makes the widget visible.
 * It will recusively go down making children widgets also visible, no matter 
 * what @recusiveinprogress is set to.  Use ui_showwidget if you dont want 
 * recusive action. On the first pass, @recusiveinprogress is %FALSE, then 
 * on recusive children it is set to %TRUE, this is so it knows that there 
 * visiblity should be only set if there visiblity is set to true in the 
 * xml defination.
 * 
 * Returns: Nothing
 */
void
bond_show(BondObject * bobj, gboolean recusiveinprogress)
	{
	GList *walk;

	g_assert(bobj);
	g_assert(bobj->oprnode);
	/* If not recusively just show the individual widget and leave 
	 * function */
	if (recusiveinprogress == FALSE)
		{
		/* wont show something that doesn't exist, ignore 
		 * uigridcolumn type. */
		if (bobj->widget && bobj->widgettype != UiGridColumn)
			ui_showwidget(bobj->bond->uixml, bobj->widget);
		}
	else
		if (bobj->oprnode->defaultvisible == TRUE
		        && (bobj->widgettype == UiGrid
		            || bobj->widgettype == UiForm 
			    || bobj->widgettype == UiFrame))
			/* wont show something that doesn't exist, ignore 
			 * uigridcolumn type. */
			if (bobj->widget && bobj->widgettype != UiGridColumn)
				ui_showwidget(bobj->bond->uixml, bobj->widget);

	for (walk = g_list_first(bobj->children); walk != NULL; 
			walk = walk->next)
		bond_show(walk->data, TRUE);
	}

/**
 * bond_close:
 * @bobj: Bond Object to close.
 * @save: If the form is to be saved on closing.
 *
 * Hide a bond object from view, save its contents and remove its registrations.
 * It will only hide the widget if the widget is of type UiForm or UiFrame or 
 * UiGrid. It also calls bond_free() to free all the memory assocated with 
 * this object, so all data specified in the widget is closed.  Any other 
 * windows that depend on this form will also be closed, and there contents 
 * saved.
 * 
 * Returns: Negative if an error occured while calling bond_save()
 */
gint
bond_close(BondObject * bobj, gboolean save)
	{
	GList *walk;
	gint retval = 0;
	BondObject *childbobj;

	mem_verify(bobj);
	/* Save if i am suppose to. It woudl be done recusively though all 
	 * children in the form */
	if (save)
		retval = bond_save(bobj, TRUE, TRUE);

	if (bobj->widget && (bobj->widgettype == UiGrid ||
	                     bobj->widgettype == UiForm
	                     || bobj->widgettype == UiFrame))
		ui_hidewidget(bobj->bond->uixml, bobj->widget);

	/* check for children windows to save data on also */
	if (bobj->obj && bobj->oprnode->childrenwindow)
		{
		/* Go around in a loop and save the contents of every window 
		 * we like with the same parameters we got */
		for (walk = g_list_first(bobj->oprnode->childrenwindow);
		        walk != NULL; walk = walk->next)
			{
			/* m.essage("trying to find %s",walk->data); */
			childbobj = bond_getbyname(bobj->bond, 
					(gchar *) walk->data);
			/* if it hasn't been drawn forget about it */
			if (childbobj == NULL)
				continue;
			if (ui_isvisible(bobj->bond->uixml, childbobj->widget) 
					== FALSE)
				continue;
			bond_close(childbobj, FALSE);
			}
		}

	/* Call bond_free(); */
	g_hash_table_remove(bobj->bond->globalbobj, bobj->widgetname);

	return retval;
	}

/* ============================================================
 * User interface control feed and life.
 * ============================================================
 */

/* Read from database the value for @bobj and save it to @value. */
static gint
bond_paint_readdb(BondObject * bobj, gchar ** value)
	{
	Registration *reg;
	gint retval = 1;
	GList *walk;

	*value = NULL;
	for (walk = g_list_first(bobj->functions); walk != NULL; 
			walk = walk->next)
		{
		warningmsg("Code me. please. now . damit. or else. AGRrrI  "
		           "SAID CODE ME");
		/* A real question, do i run this first or later when user 
		 * enters the box. */
		}
	if (bobj->oprnode->field != NULL)
		{
		registration_remove_bybobj(bobj);

		/* m.essage("reading database value from %s.%s for"
		   "%s",bobj->oprnode->table,bobj->oprnode->field, bobj->widgetname); */
		if (!bobj->obj)
			{
			errormsg("No database object associated with %s, "
			         "aborting.", bobj->widgetname);
			return -1;
			}
		if (db_numrecord(bobj->obj) <= 0 || db_isdeadobject(bobj->obj))
			{
			debuglog(100, "No current record so can not read field "
			         "%s.%s. %d %s",
			         bobj->oprnode->table, bobj->oprnode->field,
			         bobj->obj->num, bobj->obj->basetable);
			return -2;
			}
		if (db_getvalue(bobj->obj, bobj->oprnode->field,
		                bobj->oprnode->table, value) < 0)
			{
			warningmsg("Unexpected return value while trying"
			           " to extract field %s, but returned %s",
			           bobj->oprnode->field, *value);
			return -3;
			}
		
		retval = 0;
		/* m.essage("Got value %s from %s at row %d",*value,
		   bobj->oprnode->field,db_currentrow(bobj->obj)); */
		/* message("registering %s.%s", bobj->oprnode->table,
		   bobj->oprnode->field);  */
		/* Register yourself */
		reg = registration_add(bobj, bobj->obj, bobj->oprnode->table,
		                       bobj->oprnode->field, FALSE);
		if (reg != NULL)
			bobj->registrationlist =
			    g_list_prepend(bobj->registrationlist, reg);
		}
	return retval;
	}

gint
bond_save_writedb(BondObject * bobj, gchar * value, gboolean mark)
	{
	gint retval = 1;

	if (bobj->oprnode->field != NULL && bobj->oprnode->readonly == FALSE)
		{
		/* m.essage("Writing value %s to %s at row %d", value,
		   bobj->oprnode->field, db_currentrow(bobj->obj));  */
		if (db_setvalue(bobj->obj, bobj->oprnode->field,
		                bobj->oprnode->table, value) < 0)
			{
			warningmsg("Unexpected return value while trying to "
			           "write field %s, value %s",
			           bobj->oprnode->field, value);
			return -1;
			}
		retval = 0;
		}
	else if (bobj->oprnode->field == NULL)
		errormsg("This has no field to write back to");
	return retval;
	}

static gint
bond_save_masteroftable(BondObject * bobj)
	{
	/* Field stuff */
	gchar *key, *value;
	DbField *field;

	/* for walking though the registrations */
	BondObject *regbobj;
	Registration *reg;
	GList *walk, *reglist;

	if (!bobj->obj)
		return -1;
	walk = db_fieldschanged(bobj->obj);
	message("Doing master of table modify from %s. %d items",
	        bobj->widgetname, g_list_length(bobj->obj->fieldschanged));
	if (!walk)
		return -2;
	/* Get a list of any fields that have changed after a save occurres and
	   then proceed to change there values */
	for (; walk != NULL; walk = walk->next)
		{
		field = walk->data;
		key = registration_getkey(bobj->obj, field->table, field->field);

		reglist = registration_getbykey(bobj->bond->globalreg, key);
		if (g_list_length(reglist) != 1)
			{
			warningmsg("Getting key %s didn't return the expected "
				"%d number of records.", key, 
				g_list_length(reglist));
			mem_free(key);
			continue;
			}
		reg = reglist->data;
		regbobj = reg->registraint;
		if (bond_paint_readdb(regbobj, &value))
			{
			warningmsg("Faiiled to read data from %s", 
					regbobj->widgetname);
			mem_free(key);
			continue;
			}
		/* m.essage("changing widget %s", regbobj->widgetname); */
		if (regbobj->widgettype == UiGrid)
			{
			biglist_updatecell(regbobj, reg->row, reg->tablename,
			                   reg->fieldname, value);
			}
		else
			{
			bond_paint_setui(regbobj, value, TRUE);
			}
		mem_free(key);
		}
	return 0;
	}

/**
 * bond_paint_setui:
 * @bobj: Bond object
 * @value: String value to be set into the widget in @bobj.
 * @literefresh:
 * 
 * Sends the call to the user interface to display @value in the widget 
 * contained in @bobj.  This handles all the different widgets, some of the 
 * more complex widgets are passed to different functions like UiDropDown 
 * and UiGrid.
 * 
 */
void
bond_paint_setui(BondObject * bobj, gchar * value, gboolean literefresh)
	{
	switch (bobj->widgettype)
		{
	case UiNull:
	case UiGridColumn:
	case UiGeneral:
		break;
	case UiLabel:
	case UiTextBox:
	case UiCheckButton:
	case UiToggleButton:
	case UiSpinButton:
	case UiRadioButton:
	case UiEntryBox:
		/* debuglog(70,"Showing widget %s with %s", bobj->widgetname,
		   value); */
		ui_setvalue(bobj->bond->uixml, bobj->widget, value);
		if (bobj->editable == TRUE && bobj->oprnode->readonly == TRUE)
			ui_seteditable(bobj->bond->uixml, bobj->widget, FALSE);
		else
			ui_seteditable(bobj->bond->uixml, bobj->widget, bobj->editable);
		break;
	case UiDropDown:
		/* debuglog(70,"Showing dropdown box %s with %s",
		   bobj->widgetname,value); */
		dropdown_draw(bobj, value, literefresh);
		/* Set if the widget is editable or not */
		if (bobj->editable == TRUE && bobj->oprnode->readonly == TRUE)
			ui_seteditable(bobj->bond->uixml, bobj->widget, FALSE);
		else
			ui_seteditable(bobj->bond->uixml, bobj->widget, bobj->editable);
		if (!bobj->editable && dropdown_isdynamic(bobj))
			{
			ui_setvalue(bobj->bond->uixml, bobj->widget, NULL);
			ui_dropdown_clear(bobj->bond->uixml, bobj->widget);
			}
		break;
	case UiGrid:
		/* debuglog(70,"Showing a big list called %s", bobj->widgetname); */
		biglist_draw(bobj, literefresh);
		break;
	default:
		warningmsg("Unknown widget type on object %s", bobj->widgetname);
		break;
		}
	}

/**
 * bond_save_getui:
 * @bobj: Bond object to get the values from
 * @value: The return value thats saved in.
 * 
 * @bobj must have a widget associated with it.  Based on what type of 
 * widget @bobj is, it gets whatever string value is stored. It will access 
 * the ui library to retrieve the string and place it in @value.  Value is 
 * set to %NULL if there is no avaliable data.
 * 
 * Returns: Nothing at all.
 */
void
bond_save_getui(BondObject * bobj, gchar ** value)
	{
	*value = NULL;
	switch (bobj->widgettype)
		{
	case UiNull:
		break;
	case UiGeneral:
	case UiLabel:
	case UiTextBox:
	case UiCheckButton:
	case UiToggleButton:
	case UiRadioButton:
	case UiSpinButton:
	case UiEntryBox:
		debuglog(70, "Reading widget %s", bobj->widgetname);
		ui_getvalue(bobj->bond->uixml, bobj->widget, value);
		break;
	case UiDropDown:
		debuglog(70, "Reading dropdown box %s", bobj->widgetname);
		ui_getvalue(bobj->bond->uixml, bobj->widget, value);
		*value = dropdown_getkey_byvalue(bobj, *value);
		break;
	case UiGrid:
		break;
	default:
		warningmsg("Unknown widget type on object %s", bobj->widgetname);
		break;
		}
	}

/**
 * bond_paint:
 * @bobj: Bond object which contains a widget.
 * @literefresh: If %TRUE the widget will be fully redrawn, else it will
 * be only lightly updated.
 * @recusive: If it is to recusively go though the children.
 * 
 * This function will go though and extract a database value for a
 * field, and then places that value into the widget.  Signals will be
 * set up on the widget for handling when it changes.
 * 
 * Returns: Non-zero on error.
 */
gint
bond_paint(BondObject * bobj, gboolean literefresh, gboolean recursive)
	{
	gint retval = 0;
	GList *walk;
	gchar *value = NULL;

	g_assert(bobj);

	bobj->ready = FALSE;
	/* if ordered to cache a objectsrc then always do a literefresh */
	if (bobj->oprnode->dbobj && !literefresh)
		if (bobj->oprnode->dbobj->cache == TRUE)
			literefresh = TRUE;
	if (bobj->oprnode->triggerlist)
	     bondtrigger_checkevent(bobj,EventBeforePaint);
	if (bond_isfieldactive(bobj))
		{
		/* readdb, if can't disable widget */
		if (bond_paint_readdb(bobj, &value) < 0 && bobj->widget)
			{
			bobj->editable = FALSE;
			/* m.essage("editable false"); */
			}
		else if (bobj->widget)
			bobj->editable = TRUE;
		bond_paint_setui(bobj, value, literefresh);
		}
	/* Attach signals to the data. */
	postit_initsignals(bobj);

	if (recursive)
		for (walk = g_list_first(bobj->children); walk != NULL; walk = walk->next)
			retval -= bond_paint(walk->data, literefresh, recursive);
	bobj->ready = TRUE;

	if (bobj->oprnode->triggerlist)
	     bondtrigger_checkevent(bobj,EventAfterPaint);
	   
	return retval;
	}

/**
 * bond_save:
 * @bobj: Bond object 
 * @mark: If set to %TRUE, it will information to the database only if the 
 * user has entered data in the field.
 * @recusive: Whether or not to go recusively though children widgets 
 * saving contents.
 * 
 * 
 * 
 * Returns: Non-zero on error but that doesnt really happen.
 */
gint
bond_save(BondObject * bobj, gboolean mark, gboolean recursive)
	{
	GList *walk;
	gchar *value;
	BondObject *childbobj;

	/* Will exit prematurely if the assertion fails */
	mem_verify(bobj);
	/* m.essage("doing a save on %s",bobj->widgetname); */
	/* Save information to the database only if the user has entered data in the
	   field, the data is different to what is already stored and we are
	   currently painting this object */
	if (bobj->fieldactive == TRUE && bobj->changed == TRUE &&
	        bobj->ready == TRUE)
		{
		if (db_isdeadobject(bobj->obj) == TRUE)
			debuglog(60, "Ignoring save on %s due to empty dbobject.",
			         bobj->widgetname);
		else
			{
			debuglog(90, "Saving widget %s", bobj->widgetname);
			bond_save_getui(bobj, &value);
			bond_save_writedb(bobj, value, mark);
			bond_save_masteroftable(bobj);
			bobj->changed = FALSE;
			}
		}
	/* else m.essage("%d %d %d
	   %s",bobj->fieldactive,bobj->changed,bobj->ready,bobj->oprnode->field); */
	/* Save any children for this object that exist. */
	if (recursive)
		for (walk = g_list_first(bobj->children); walk != NULL; walk = walk->next)
			bond_save(walk->data, mark, recursive);

	/* check for children windows to save data on also. This is if you have
	   <childwindow> set on a widget. It branchs out and saves these also. */
	if (bobj->obj && bobj->oprnode->childrenwindow)
		{
		/* Go around in a loop and save the contents of every window we like
		   with the same parameters we got */
		for (walk = g_list_first(bobj->oprnode->childrenwindow);
		        walk != NULL; walk = walk->next)
			{
			/* m.essage("trying to find %s",walk->data); */
			childbobj = bond_getbyname(bobj->bond, (gchar *) walk->data);
			/* if it hasn't been drawn forget about it */
			if (childbobj == NULL)
				continue;
			/* if it isn't displayed dont save it */
			if (ui_isvisible(bobj->bond->uixml, childbobj->widget) 
					== FALSE)
				continue;
			bond_save(childbobj, mark, recursive);
			}
		}

	return 0;
	}

/**
 * bond_repaint_bychange:
 * @bobj:
 * @paintby:
 * @row:
 * @tablename:
 * @fieldname:
 * 
 * Sometimes a change in one widget can effect the values of another widget
 * if they share the same data. For example if a entry widget shares data with
 * a grid then it will update the other not updated widget with the current
 * data. This is all driven by registrations of database objects. 
 * 
 * Returns: Nothing at all.
 */
void
bond_repaint_bychange(BondObject * bobj, BondObject * paintby, gulong row,
                      gchar * tablename, gchar * fieldname, gchar *newtext)
	{
	gchar *value;
	if (paintby == bobj)
		{
		warningmsg("You can't repaint yourself %s. leaving", 
				bobj->widgetname);
		return ;
		}
	if (newtext == NULL)
		{
		if (paintby->fieldactive == TRUE || paintby->active == TRUE)
			{
			bond_save_getui(paintby, &value);
			}
		else
			return ;
		}
	else
		value = newtext;
	
	mem_verify(bobj);
	mem_verify(paintby);
	debuglog(40, "Read value %s from %s. Putting it into another "
	         "widget.%s", value, paintby->widgetname, bobj->widgetname);
	if (bobj->widgettype == UiGridColumn)
		{
		biglist_updatecell(bobj, row, tablename, fieldname, value);
		}
	else
		{
		bond_paint_setui(bobj, value, TRUE);
		}
	}

/**
 * bond_refresh:
 * @bobj: Bond object.
 * 
 * Re-paint @bobj with current data using bond_paint().
 *
 * Returns: Non-zero on error.
 */
gint
bond_refresh(BondObject * bobj)
	{
	return bond_paint(bobj, FALSE, TRUE);
	}

/**
 * bond_ischanged:
 * @bobj: Bond object to check.
 * 
 * Figure out if the @bobj has changed and should be saved.
 * 
 * Returns: %TRUE or %FALSE.
 */
gboolean
bond_ischanged(BondObject * bobj)
	{
	return bobj->changed;
	}

/**
 * bond_setchanged:
 * @bobj: Bond object to set.
 * 
 * Sets the state of @bobj to say that it has been changed by user actions.
 * This occurs when someone types a value in the widget and is called by 
 * postit.c.  This is also called by grid editing.
 * 
 * Returns: Non-zero on error.
 */
gint
bond_setchanged(BondObject * bobj, gboolean changed, gchar *newtext)
	{
	GList *reglist, *walk;
	Registration *reg = NULL;

	/* mark the change */
	if (bobj->oprnode->readonly)
		{
		debuglog(80, "Field %s.%s is marked as readonly, I can't "
		         "save changes.", bobj->oprnode->table, 
			 bobj->oprnode->field);
		return -1;
		}

	bobj->changed = changed;
	/* Handle any other objects currently drawn that may contain the same
	 * information */
	reglist = registration_check(bobj);

	/* debugging output */
	if (reglist)
		message("%d items registered under key %s", 
				g_list_length(reglist),
		        ((Registration *) bobj->registrationlist->data)->key);
	
	for (walk = g_list_first(reglist); walk != NULL; walk = walk->next)
		{
		reg = walk->data;
		mem_verify(reg);
		/* registration_debug(reg); */
		if (reg->registraint != bobj)
			{
			debuglog(40,
			         "Widget change has caused an update "
				 "somewhere else.");
			bond_repaint_bychange(reg->registraint, bobj, reg->row,
			                      reg->tablename, reg->fieldname,
					      newtext);
			}
		}
	/* m.essage("setting watchtower change on %s",bobj->widgetname); */
	watchtower_change(bobj);

	return 0;
	}

/* ============================================================
 * Bond object spawning and searching.
 * ============================================================
 */

static BondObject *
bond_spawn_rec(BondObject * parentbobj, BondRelation * parentbrel,
               gboolean spawnnew)
	{
	BondObject *bobj = NULL;
	OperationsNode *node;
	BondRelation *brel = parentbrel;
	GList *walk;

	for (walk = g_list_first(parentbobj->oprnode->childrennode); walk != NULL;
	        walk = walk->next)
		{
		node = walk->data;
		if (spawnnew == TRUE)
			{
			errormsg("Hey write me ");
			/* Write me, handle code for spawns of multi instances */
			}
		else
			{
			g_hash_table_remove(parentbobj->bond->globalbobj,
			                    node->widgetname);
			}
		bobj = bond_create(parentbobj->bond, node);
		bobj->parent = parentbobj;
		/* if (parentbrel != NULL) m.essage("parentbrel is set in
		   %s",bobj->widgetname); if (bobj->oprnode->dbobj != NULL)
		   m.essage("node->dbobj is set"); */
		if (parentbrel != NULL && bobj->oprnode->dbobj != NULL)
			{
			bobj->brel = parentbrel;
			bobj->brel->refcounter++;
			brel = NULL;
			parentbrel = NULL;
			}
		/* if (brel != NULL) m.essage("running spawn with brel set"); if
		   (bobj->brel != NULL) m.essage("brel set in bobj"); */

		bondrelation_anaylise(bobj);
		bond_registerobj(bobj);
		parentbobj->children = g_list_prepend(parentbobj->children, bobj);
		bond_spawn_rec(bobj, brel, spawnnew);
		/* m.essage("adding child %s to %s "
		   "%d",bobj->widgetname,parentbobj->widgetname,
		   g_list_length(parentbobj->children)); */
		}
	if (parentbobj != NULL)
		parentbobj->children = g_list_reverse(parentbobj->children);
	return bobj;
	}

/**
 * bond_spawn:
 * @bond: Global bond object.
 * @parentbobj: BondObject to spawn from.
 * @widgetname: Name of widget that bondobject belongs to.
 * @brel: Bond relation, instructions on how this document relates to @parentbobj.
 * @spawnnew: Normally %FALSE,  if it is to spawn another copy of the bond object.
 * 
 * This function will generate a new bond object based on another bond object.
 * It will load the object as a child to @parentbobj.  An OperationsNode is
 * found from @widgetname giving the definiation found in the bond xml document
 * for how this document is created.
 * 
 * 
 * Returns: A newly created BondObject*.  Free by removing item from the hash 
 * table in @bond.
 */
BondObject *
bond_spawn(Bond * bond, BondObject * parentbobj, gchar * widgetname,
           BondRelation * brel, gboolean spawnnew)
	{
	BondRelation *childbrel;
	BondObject *bobj = NULL;
	OperationsNode *oprnode;

	/* Initializing n testing stuff */
	if (widgetname == NULL)
		{
		errormsg("Widgetname are NULL. Can not proceed with the spawn.");
		return NULL;
		}
	if (spawnnew == TRUE)
		{
		warningmsg("This code has not been done yet. Be kind enough to\n"
		           "write it for me");
		/* Write me, handle code for spawns of multi instances */
		}
	else
		g_hash_table_remove(bond->globalbobj, widgetname);
	childbrel = brel;
	oprnode = bondxml_getnode(bond, widgetname);
	if (oprnode == NULL)
		{
		errormsg("Can not find widget %s, aborting spawn", widgetname);
		return NULL;
		}
	/* If bond relation is set for loading the bondobject by then use it */
	if (brel != NULL)
		{
		if (brel->dbobj == NULL)
			brel->dbobj = oprnode->dbobj;
		brel->refcounter++;
		if (oprnode->dbobj != NULL)
			childbrel = NULL;
		}
	/* Create the bond object and database objects to being with */
	bobj = bond_create(bond, oprnode);
	bobj->brel = brel;
	bobj->parent = parentbobj;
	bondrelation_anaylise(bobj);
	bond_registerobj(bobj);
	/* Spawn children now */
	bond_spawn_rec(bobj, childbrel, spawnnew);
	if (bobj->parent != NULL)
		{
		g_assert(parentbobj);
		parentbobj->children = g_list_append(parentbobj->children, bobj);
		}
	return bobj;
	}

static gboolean
bond_isobjparent(BondObject * bobj)
	{
	if (!bobj->obj)
		return FALSE;
	if (bobj->parent != NULL)
		if (bobj->obj == bobj->parent->obj)
			return FALSE;
	return TRUE;
	}

/* ============================================================
 * Bond object registration handling - similar to watchtower.
 * ============================================================
 */

static gint
bond_registerobj_obj(Object * obj, BondObject * bobj)
	{
	gboolean insertnew = FALSE;
	GList *objlist;

	/* m.essage("Registering %s on %s",obj->basetable,bobj->widgetname); */
	/* add the object into the registery, under its object pointer. */
	objlist = g_hash_table_lookup(bobj->bond->globalbobj_byobj, obj);
	if (objlist == NULL)
		insertnew = TRUE;
	objlist = g_list_append(objlist, bobj);
	if (insertnew == TRUE)
		g_hash_table_insert(bobj->bond->globalbobj_byobj, obj, objlist);
	return 0;
	}

static gint
bond_unregisterobj_obj(Object * obj, BondObject * bobj)
	{
	GList *objlist, *newlist, *walk;

	/* look it up */
	objlist = g_hash_table_lookup(bobj->bond->globalbobj_byobj, obj);
	/* If it was all a lie and it wans't registered return nicely */
	if (objlist == NULL)
		return -1;
	/* Difficult bit, lookup returns a GList*, so we have to carefully remove
	   ourselves from this list. If this is last item on the list we have to
	   free the list and remove the hash table entry completey. */
	if (g_list_length(objlist) == 1)
		{
		g_hash_table_remove(bobj->bond->globalbobj_byobj, obj);
		return 0;
		}
	/* If we are at the head of hte list then the pointer address will change
	   for the list so we have to rebuild the list from scratch and re-insert
	   it into the hash table */
	if (objlist->data == bobj)
		{
		newlist = NULL;
		objlist = objlist->next;

		for (walk = objlist; walk != NULL; walk = walk->next)
			newlist = g_list_prepend(newlist, walk->data);

		g_hash_table_remove(bobj->bond->globalbobj_byobj, obj);
		g_hash_table_insert(bobj->bond->globalbobj_byobj, obj, newlist);
		}
	else
		newlist = g_list_remove(objlist, bobj);
	return 0;
	}

/**
 * bond_registerobj:
 * @bobj: Bond object.
 *
 * Sometimes more than one bond object can share the same database object, eg
 * a biglist and a bunch of widgets that point to the row selected in the list.
 * Objects are registered so that we know what owns what.  
 *
 * Returns: non-zero on error.
 */
gint
bond_registerobj(BondObject * bobj)
	{
	GList *walk;

	/* check if parent obj is same as this obj. Ie this is defined as
	   <dbobject> is bond xml doc */
	if (!bond_isobjparent(bobj))
		return -1;
	/* If a bondrelation exists. Then for every parent object in it, register
	   as belonging to us. This is so if the parent is closed, then so are we
	   because we need the parent. Bondrelations have a GList* of bond objects
	   they might derive from. */
	if (bobj->brel)
		if (bobj->brel->parentobj &&
		        (Object *) bobj->brel->parentobj->data != bobj->obj)
			{
			for (walk = g_list_first(bobj->brel->parentobj);
			        walk != NULL; walk = walk->next)
				bond_registerobj_obj(walk->data, bobj);
			debuglog(80, "Another object is used to load this so "
			         "i am registering this as well; %s", bobj->widgetname);
			}
	/* Register ourselves */
	return bond_registerobj_obj(bobj->obj, bobj);
	}

/**
 * bond_unregisterobj:
 * @bobj: Bond Object.
 * 
 * Unregisters a object that has previously been register. This isn't to do with
 * freeing a bond object but meerly stating that the database object is no 
 * long in use and should not be mointored.  
 * 
 * Returns: non-zero on error.
 */
gint
bond_unregisterobj(BondObject * bobj)
	{
	GList *walk;

	/* abort if it isn't a parent type bobj */
	if (!bond_isobjparent(bobj))
		return -2;
	if (bobj->brel)
		if (bobj->brel->parentobj)
			{
			for (walk = g_list_first(bobj->brel->parentobj); walk != NULL;
			        walk = walk->next)
				bond_unregisterobj_obj(walk->data, bobj);
			debuglog(80, "Unregistering %s loadby objects", bobj->widgetname);
			}
	return bond_unregisterobj_obj(bobj->obj, bobj);
	}

/**
 * bond_registerobjchange:
 * @parentbobj: Bond object that has changed.
 * @objchange: Function to run once the @parentobject has been saved.
 * @literefresh: 
 *
 * Informs the registor that a widget/BondObject/DbObj has changed.  It finds
 * all bond objects that use @obj, and redraws them all.  It only finds objects
 * which are the parent, ignores all widgets that just inherint from there
 * parents.
 * 
 * Returns: non-zero on error.
 */
gint
bond_registerobjchange(BondObject * parentbobj, gint(*objchange) (BondObject *),
                       gboolean literefresh)
	{
	gint retval = 0;
	GList *objlist, *walk, *copylist;
	BondObject *bobj = NULL;
	gboolean tmp = FALSE;
	Object *origobj;

	/* (debugging) If no parent object exists, terminate program */
	g_assert(parentbobj);
	/* m.essage("BondObject %s has changed, I'll save it.",
	   parentbobj->widgetname); */

	/* Abort if we have no records yet. */
	if (parentbobj->bond->globalbobj_byobj == NULL)
		return -1;
	if (parentbobj->obj == NULL)
		return -2;
	/* Get a list of all the widgets that share this data */
	objlist = g_hash_table_lookup(parentbobj->bond->globalbobj_byobj,
	                              parentbobj->obj);
	/* ==================================================== */
	/* Save all forms that are open that are going to be effected. */
	/* The list is going to change as we save stuff so we need a copy */
	copylist = g_list_copy(objlist);
	for (walk = g_list_first(objlist); walk != NULL; walk = walk->next)
		{
		bobj = walk->data;
		debuglog(80,"%s affected by object change in %s", 
				bobj->widgetname, parentbobj->widgetname); 
		/* if the parent db object is the same as this db object then 
		 * save */
		if (bobj->obj == parentbobj->obj)
			bond_save(bobj, TRUE, FALSE);
		}
	origobj = parentbobj->obj;
	/* ==================================================== */
	/* Handle the dead objects ... This is a in the case where you have a 
	 * form with no current data in it that has a refresh or new database 
	 * call on it to give it data, all the widgets need to be 
	 * reactiviated. */
	if (db_isdeadobject(parentbobj->obj))
		{
		if (parentbobj->brel == NULL)
			{
			errormsg("No bond relation specified.");
			g_list_free(copylist);
			return -1;
			}

		debuglog(60, "Object in %s is dead so i'm going to create "
		         "it first.", parentbobj->widgetname);
		/* Figure out again what are database data is */
		bondrelation_anaylise(parentbobj);
		}

	g_assert(parentbobj->obj);
	/* ==================================================== */
	/* Go though and change all the database objects now. This is the 
	 * crunch point, this work is done by the function parsed in. */
	debuglog(50,
	         "Object %s has changed so I'm going to update %d items "
	         "that shares its object", parentbobj->widgetname,
		 g_list_length(copylist));
	
	/* Run the function specified */
	retval = objchange(parentbobj);
	debuglog(50, "Finished running change function");
	
	/* Handle failure nicely */
	if (parentbobj->obj == NULL)
		{
		tmp = TRUE;
		errormsg("obj is null. sooo we should update nicely");
		}
	else
		objlist = g_hash_table_lookup(
			parentbobj->bond->globalbobj_byobj, parentbobj->obj);

	/* ==================================================== */
	/* Redraw everything that was effected. */
	for (walk = g_list_first(copylist); walk != NULL; walk = walk->next)
		{
		bobj = walk->data;
		/* We dont want to redraw ourselves - thats just confusing */
		/* if (bobj == parentbobj)
			continue; */
		debuglog(80, "Repainting %s because of change of %s",
		         bobj->widgetname, parentbobj->widgetname);
		/* If in another dbobjectsrc and this is the parent dbobject 
		 * then refresh the bobj->obj so its uptodate. This is for 
		 * form->form and loadby relationships. */
		/* If the object you are refreshing has data inside it redraw 
		 * its children also. This will call this function on itself, 
		 * if it isn't the same data source. */
		if (bobj->brel && bobj->obj != parentbobj->obj)
			bond_registerobjchange(bobj, bondrelation_refresh, 
					FALSE);
		else
			bond_paint(bobj, literefresh, FALSE);
		}
	/* Free the copy of a list we dont need anymore and get out of here */
	g_list_free(copylist);
	return retval;
	}
