/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <glib.h>

#include "db.h"
#include "bc.h"
#include "ui.h"

#include "bondapi.h"
#include "bondmain.h"
#include "biglists.h"
#include "bond.h"
#include "bondrelation.h"

/**
 * Internal Functions
 *
 * These bond_api_*_internal functions are used to carry out the actual
 * api action and are called from elsewhere in bond as pointers to
 * functions.
 */
static gint
bond_api_add_internal(BondObject * bobj)
	{
	if (bobj->obj)
		{
		if (db_isrecordset(bobj->obj) == FALSE
			&& bobj->brel)
			{
			debuglog(30,"Refreshing widget %s because object is "
				"dead.",bobj->widgetname);
			bondrelation_refresh(bobj);
			}
		return db_add(bobj->obj);
		}
	return -1;
	}

static gint
bond_api_new_internal(BondObject * bobj)
	{
	if (!bobj->obj)
		return -1;
	db_clearobject(bobj->obj);
	return db_add(bobj->obj);
	}

static gint
bond_api_loadall_internal(BondObject * bobj)
	{
	if (!bobj->obj)
		return -1;
	db_loadall(bobj->obj);
	return 0;
	}

static gint
bond_api_next_internal(BondObject * bobj)
	{
	if (!bobj->obj)
		return -1;
	return db_movenext(bobj->obj);
	}

static gint
bond_api_prev_internal(BondObject * bobj)
	{
	gint currentrow;

	if (!bobj->obj)
		return -2;

	currentrow = db_currentrow(bobj->obj);

	if (currentrow - 1 < 0)
		return -1;
	return db_moveto(bobj->obj, currentrow - 1);
	}

static gint
bond_api_first_internal(BondObject * bobj)
	{
	gint numrow;

	if (!bobj->obj)
		return -2;

	numrow = db_numrecord(bobj->obj);

	if (numrow <= 0)
		return -1;
	return db_moveto(bobj->obj, 0);
	}

static gint
bond_api_last_internal(BondObject * bobj)
	{
	if (!bobj->obj)
		return -2;
	return db_movelast(bobj->obj);
	}

static gint
bond_api_refresh_internal(BondObject * bobj)
	{
	if (!bobj->obj)
		return -2;
	return bondrelation_refresh(bobj);
	}

static gint
bond_api_clear_internal(BondObject * bobj)
	{
	if (!bobj->obj)
		return -2;
	return db_obj_clear(bobj->obj);
	}

static gint
bond_api_delete_internal(BondObject * bobj)
	{
	if (!bobj->obj)
		return -2;
	message("Doing delete on %s",bobj->obj->basetable);
	return db_deleteobject(bobj->obj);
	}

static gint
bond_api_brelrefresh_internal(BondObject * bobj)
	{
	if (bobj->brel)
		return bondrelation_anaylise(bobj);
	return 0;
	}

static gint
bond_api_abortsave_internal(BondObject * bobj)
	{
	GList *walk;
	bobj->changed = FALSE;
	for (walk=g_list_first(bobj->children);walk!=NULL;walk=walk->next)
		{
		bond_api_abortsave_internal(walk->data);
		}		
	return 0;	
	}

/**
 * bond_api_delete:
 * 
 * Deletes bondObject widget associated with @widgetname. (bond1 ->
 * Probably should use bond_specifyclass() first.  This will also delete
 * any objects that refer to the selected object as well. So use with
 * care.)
 *
 * Returns: Non-zero on error.
 */
gint
bond_api_delete(Bond * bond, gchar * widgetname)
	{
	gint retval;
	gchar *lasterror, *message;
	BondObject *bobj;

	bobj = bond_getbyname(bond, widgetname);
	debug_output("API Call: Delete %s\n", widgetname);
	if (!bobj)
		{
		errormsg("BondObject for %s can not be found. Make sure "
			"widget is spawned and open first.", widgetname);
		return -1;
		}
	retval = bond_registerobjchange(bobj, bond_api_delete_internal, FALSE);
	/* Display any sql errors if they happen */
	if ((lasterror	= db_getlasterror()) != NULL)
		{
		message = mem_strdup_printf("Deletion failed on %s\n%s",
				widgetname,lasterror);
		ui_display_message(bond->uixml,message,"Deletion failed!",0);
		mem_free(message);
		retval = -2;
		}

	return retval;
	}

/**
 * bond_api_add:
 * @bond: Global bond object.
 * @widgetname: Widget which has a bondobject associated that you wish to add 
 * a record to.
 * 
 * Adds widget associated with @widgetname, only if the widget exists
 * and it has an associated basetable. (bond1 -> Prepares to insert new
 * data into the database.  If the form is not already visible it will
 * be made visible and then a new record fresh to be entered into will
 * be there.  If the form Is already visible it will be blanked and then
 * the initial defaults put in.)
 *
 * Returns: Non-zero on error.
 */
gint
bond_api_add(Bond * bond, gchar * widgetname)
	{
	gboolean repaint = FALSE;
	gint retval;
	BondObject *bobj;

	/* Get the bond object with the inputted widget name */
	bobj = bond_getbyname(bond, widgetname);
	debug_output("API Call: Add %s\n", widgetname);

	if (!bobj)
		{
		bobj = bond_spawn(bond, NULL, widgetname, NULL, FALSE);
		if (!bobj)
			{
			errormsg("BondObject for %s can not be found. Make "
			"sure widget is spawned and open first.", widgetname);
			return -1;
			}
		else
			repaint = TRUE;
		bond_paint(bobj, FALSE, TRUE);
		return -1;
		}
	if (bobj->obj && !bobj->obj->basetable)
		{
		errormsg("DbObject %s has no base table in widget %s", widgetname);
		return -2;
		}
	if (!bobj->obj)
		{
		errormsg("No dbobject is specified on %s but a database call"
			 " has been made against it.",widgetname);
		return -3;
		}
	   
	retval = bond_registerobjchange(bobj, bond_api_add_internal, FALSE);
	/* if (repaint==TRUE)
	     bond_paint(bobj,FALSE,TRUE); */ 
	return retval;
	}

/**
 * bond_api_new:
 * 
 * Prepares to insert new data into the database.  If the form is not
 * already visible it will be made visible and then a new record fresh
 * to be entered into will be there.  If the form is already visible it
 * will be blanked and then the initial defaults put in.
 *
 * Returns: Non-zero on error.
 */
gint
bond_api_new(Bond * bond, gchar * widgetname)
	{
	BondObject *bobj;

	bobj = bond_getbyname(bond, widgetname);
	debug_output("API Call: New %s\n", widgetname);

	if (!bobj)
		{
		errormsg("BondObject for %s cannot be found. Make sure widget is"
		         " spawned and open first.", widgetname);
		return -1;
		}
	if (bobj->obj && !bobj->obj->basetable)
		{
		errormsg("Db Object %s has no base table in widget %s", widgetname);
		return -2;
		}

	return bond_registerobjchange(bobj, bond_api_new_internal, FALSE);
	}

/**
 * bond_api_loadall:
 *
 * Loads all the records in the table which is representitive of
 * @formname.
 *
 * Returns: non-zero on error. 
 */
gint
bond_api_loadall(Bond * bond, gchar * widgetname)
	{
	BondObject *bobj;

	bobj = bond_getbyname(bond, widgetname);
	debug_output("API Call: LoadAll %s\n", widgetname);
	if (!bobj)
		{
		errormsg("BondObject for %s can not be found. Make sure widget is"
		         " spawned and open first.", widgetname);
		return -1;
		}

	return bond_registerobjchange(bobj, bond_api_loadall_internal, FALSE);
	}

/**
 * bond_api_show:
 * @widgetname: Name of a widget to display. 
 *
 * Displays the wdiget @widgetname on screen.  Initialises roof with
 * correct values.  Datasource should be set with bond_specifyclass
 * first to ensure that correct data is available.
 * 
 *
 * Returns: Non-zero on error.
 */
gint
bond_api_open(Bond * bond, gchar * widgetname)
	{
	BondObject *bobj;

	g_assert(widgetname);

	bobj = bond_getbyname(bond, widgetname);
	debug_output("API Call: Open %s\n", widgetname);
	if (!bobj)
		{
		bobj = bond_spawn(bond, NULL, widgetname, NULL, FALSE);
		if (bobj)
			bond_paint(bobj, FALSE, TRUE);
		else
			{
			errormsg("BondObject for %s can not be found. Make "
				"sure widget is spawned and open first.", 
				widgetname);
			return -1;
			}
		}
	bond_show(bobj, FALSE);
	/* Bring window to the top if it is actually a window */
	if (bobj->widgettype == UiForm)
		ui_raisewindow(bobj->bond->uixml, bobj->widget);

	return 0;
	}

/**
 * bond_api_close:
 * 
 * Removes the form from screen, and also removes any child forms from the screen.
 * Note this automatically call bond_api_save() for you.
 * 
 * Returns: non-zero on failure.
 */
gint
bond_api_close(Bond * bond, gchar * widgetname)
	{
	gint retval;
	BondObject *bobj;

	g_assert(widgetname);

	debuglog(30, "API Call: Close %s (save before closing)\n", widgetname);
	bobj = bond_getbyname(bond, widgetname);
	if (!bobj)
		{
		errormsg("BondObject for %s can not be found. Make sure "
			"widget is spawned and open first.", widgetname);
		return -1;
		}
	retval = bond_close(bobj, TRUE);
	/* Quit the app if theres nothing left open. */
	postit_checkforclosedwindows(bobj);
	return retval;
	}

/**
 * bond_api_save:
 * 
 * Saves the current form to the backend. It may not be posted to the database
 * straight away, but it will at least be queued for a write.
 * 
 * 
 * Returns: Non-zero on error.
 */
gint
bond_api_save(Bond * bond, gchar * widgetname)
	{
	gint retval=0;
	gchar *lasterror, *message;
	BondObject *bobj;

	g_assert(widgetname);
	debug_output("API Call: Save %s\n", widgetname);
	bobj = bond_getbyname(bond, widgetname);
	if (!bobj)
		{
		errormsg("BondObject for %s can not be found. Make sure "
			"widget is spawned and open first.", widgetname);
		return -1;
		}
	/* recusively save every widget */
	bond_save(bobj, TRUE, TRUE);
	/* If this has a dbobject then go ahead and force a save once its all 
	 * done */
	if (bobj->obj)
		db_save(bobj->obj);
	/* Display any sql errors if they happen */
	if ((lasterror	= db_getlasterror()) != NULL)
		{
		message = mem_strdup_printf("Saving failed on %s\n%s",
				widgetname,lasterror);
		ui_display_message(bond->uixml,message,"Save failed!",0);
		mem_free(message);
		retval = -2;
		}
	return retval;
	}

/**
 * bond_api_abortsave:
 * @bond: Global bond object
 * @widgetname: Name of widget/form to abort the save on.
 *
 * Aborts any saves that are due on @widgetname.
 */
gint 
bond_api_abortsave(Bond *bond, gchar *widgetname)
	{
	BondObject *bobj;

	g_assert(widgetname);
	debug_output("API Call: Save %s\n", widgetname);
	bobj = bond_getbyname(bond, widgetname);
	if (!bobj)
		{
		errormsg("BondObject for %s can not be found. Make sure "
			"widget is spawned and open first.", widgetname);
		return -1;
		}
	return bond_api_abortsave_internal(bobj);
	}


/**
 * bond_api_refresh:
 * @bond: Global bond object.
 * @widgetname: Widget to refresh.
 * 
 * Refresh the data objects in @widgetname to say whats in the database. Any 
 * changes are saved first. @widgetname must already be displayed first.
 * 
 * Returns: non-zero on error.
 */
gint
bond_api_refresh(Bond * bond, gchar * widgetname)
	{
	BondObject *bobj;

	g_assert(widgetname);
	debug_output("API Call: Refresh %s\n", widgetname);
	bobj = bond_getbyname(bond, widgetname);
	if (!bobj)
		{
		errormsg("BondObject for %s can not be found. Make sure "
			"widget is spawned and open first.", widgetname);
		return -1;
		}
	if (!bobj->obj)
		{
		errormsg("No database object on %s so I can not refresh it.",
				widgetname);
		return -2;
		}
	return bond_registerobjchange(bobj, bond_api_refresh_internal, FALSE);
	}

/**
 * bond_api_next:
 * @formname: Name of a form which contains a database object.
 *
 * Moves to the next record in @formname, if possible.
 *
 * Returns: Non-zero on error.
 */
gint
bond_api_next(Bond * bond, gchar * widgetname)
	{
	BondObject *bobj;

	g_assert(widgetname);
	debug_output("API Call: Next %s\n", widgetname);
	bobj = bond_getbyname(bond, widgetname);
	if (!bobj)
		{
		errormsg("BondObject for %s can not be found. Make sure widget is"
		         " spawned and open first.", widgetname);
		return -1;
		}
	return bond_registerobjchange(bobj, bond_api_next_internal, TRUE);
	}

/**
 * bond_api_prev:
 * @formname: Name of a form which contains a database object.
 *
 * Moves to the previous record in @formname, if possible.
 *
 * Returns: Non-zero on error.
 */
gint
bond_api_prev(Bond * bond, gchar * widgetname)
	{
	BondObject *bobj;

	g_assert(widgetname);
	bobj = bond_getbyname(bond, widgetname);
	if (!bobj)
		{
		errormsg("BondObject for %s can not be found. Make sure widget is"
		         " spawned and open first.", widgetname);
		return -1;
		}
	return bond_registerobjchange(bobj, bond_api_prev_internal, TRUE);
	}

/**
 * bond_api_first:
 * @formname: Name of form.
 * 
 * Move to the first item in the object record set.  Once your've loaded a form, 
 * you can use this function to force it to show the first record in the roof 
 * object.
 *
 * Returns: non-zero on error.
 */
gint
bond_api_first(Bond * bond, gchar * widgetname)
	{
	BondObject *bobj;

	g_assert(widgetname);
	bobj = bond_getbyname(bond, widgetname);
	if (!bobj)
		{
		errormsg("BondObject for %s can not be found. Make sure "
			"widget is spawned and open first.", widgetname);
		return -1;
		}
	return bond_registerobjchange(bobj, bond_api_first_internal, TRUE);
	}

/**
 * bond_api_last:
 * @formname: Name of form.
 * 
 * Move to the last item in the object record set.  
 *
 * Returns: non-zero on error.
 */
gint
bond_api_last(Bond * bond, gchar * widgetname)
	{
	BondObject *bobj;

	g_assert(widgetname);
	bobj = bond_getbyname(bond, widgetname);
	if (!bobj)
		{
		errormsg("BondObject for %s can not be found. Make sure "
			"widget is spawned and open first.", widgetname);
		return -1;
		}
	return bond_registerobjchange(bobj, bond_api_last_internal, TRUE);
	}

/**
 * bond_api_clear:
 * 
 * Clears the fields of a given form.
 * 
 * Returns: non-zero on failure.
 */
gint
bond_api_clear(Bond * bond, gchar * widgetname)
	{
	BondObject *bobj;

	g_assert(widgetname);
	bobj = bond_getbyname(bond, widgetname);
	if (!bobj)
		{
		errormsg("BondObject for %s can not be found. Make sure "
			"widget is spawned and open first.", widgetname);
		return -1;
		}
	return bond_registerobjchange(bobj, bond_api_clear_internal, FALSE);
	}

gint
bond_api_clearloadby(Bond * bond, gchar * widgetname)
	{
	BondObject *bobj;

	bobj = bond_getbyname(bond, widgetname);
	/* if doesnt exist spawn the bond object. */
	if (!bobj)
		return -1;
	if (!bobj->brel)
		return -2;
	bondrelation_free(bobj->brel);
	bobj->brel = NULL;
	bondrelation_anaylise(bobj);
	return bond_api_refresh(bond, widgetname);
	}

gint
bond_api_addloadby(Bond * bond, gchar * targetwidget, gchar * loadbywidget)
	{
	GList *walk;
	BondObject *bobj, *loadby;

	g_assert(targetwidget);
	g_assert(loadbywidget);

	bobj = bond_getbyname(bond, targetwidget);
	/* if doesnt exist spawn the bond object. */
	if (bobj == NULL)
		{
		if (!bondrelation_addpending(bond, targetwidget, loadbywidget))
			return -1;
		/* Return now that i leave it for the future to handle it */
		return 0;
		}
	/* Modify a live form */
	loadby = bond_getbyname(bond, loadbywidget);
	if (!loadby)
		{
		errormsg("Can not show %s because the loadby widget %s doesn't exist",
		         targetwidget, loadbywidget);
		return -2;
		}
	if (!loadby->obj)
		{
		errormsg("Can not load by %s because it has no active database object"
		         " attached to it yet.", loadbywidget);
		return -3;
		}
	if (!bobj->brel)
		{
		errormsg("No bond relation exists on %s, please make sure it has a"
		         " correct <dbobject>.", targetwidget);
		return -3;
		}
	for (walk = g_list_first(bobj->brel->parentobj); walk != NULL;
	        walk = walk->next)
		{
		if (walk->data == loadby->obj)
			{
			debuglog(80, "Loadby object %s already added, wont add it again",
			         loadbywidget);
			return -4;
			}
		}
	/* form needs to be refreshed now */
	bobj->brel->parentobj = g_list_append(bobj->brel->parentobj, loadby->obj);
	return 0;
	}

/**
 * bond_api_sql:
 * @bond: Bond global object.
 * @widgetname: Widget where results will be assigned.
 * 
 * Execute an sql statement and send the results to @widgetname.
 * 
 * 
 * 
 */
gint
bond_api_sql(Bond * bond, gchar * widgetname, gchar * parentwidgetname,
             gchar * sqlquery)
	{
	BondRelation *brel;
	BondObject *bobj, *parentbobj = NULL;

	g_assert(widgetname);
	g_assert(sqlquery);
	/* Search to see if the widget is already displayed */
	bobj = bond_getbyname(bond, widgetname);
	debug_output("API Call: Sql %s\n", widgetname);

	if (parentwidgetname)
		parentbobj = bond_getbyname(bond, parentwidgetname);
	/* Create a relationship */
	brel = bondrelation_create(RelationSQL);
	brel->sql = sqlquery;
	/* If the form hasn't been loaded yet create it now. */
	if (bobj == NULL)
		{
		/* Spawn the object */
		bobj = bond_spawn(bond, NULL, widgetname, brel, FALSE);
		if (bobj)
			bond_paint(bobj, FALSE, TRUE);
		else
			{
			errormsg("BondObject for %s can not be spawned.", 
					widgetname);
			return -1;
			}
		}
	else
		{
		/* If a bond relationship already exists then it needs to refreshed
		   and re-run incase there are several different ways to load data *
		   into a form. */
		if (bobj->brel)
			bondrelation_free(bobj->brel);
		bobj->brel = brel;
		/* This will call bondrelation_anaylise on all the dbobjects */
		bond_registerobjchange(bobj, bond_api_brelrefresh_internal, FALSE);
		}
	bondrelation_free(brel);

	bond_show(bobj, FALSE);

	return 0;
	}

/**
 * bond_api_loadby:
 * @bond: 
 * @widgetname: 
 * @loadbyparents: 
 * 
 * Check if it has been spawned. If it has just get brel, add to
 * parent obj, list the parent obj. Call bond_refresh on the object then to
 * update it.  If not handle it another way.  In bondmain have a repository of
 * bondrelations.  In bondrelation add target widget, add loadby widgetname.
 * Have GList in Bond *bond. Everytime bondrelation_analyis is called check to
 * see if an entry exists for this widgetname. If so run it.  Modify
 * db_loadobjbyobjects to support load by itself.
 *
 * Returns: Non-zero on error.
 */
gint
bond_api_loadby(Bond * bond, gchar * widgetname, GList * loadbyparents)
	{
	BondObject *bobj;
	GList *walk;

	g_assert(widgetname);
	/* m.essage("Loadby called."); 
	m.essage("%d items in list",
	   g_list_length(loadbyparents)); */

	if (loadbyparents == NULL || g_list_length(loadbyparents) == 0)
		warningmsg("Nothing specified on loadby for %s",widgetname);
	/* bond_api_clearloadby(bond, widgetname); */
	for (walk = g_list_first(loadbyparents); walk != NULL; walk = walk->next)
		{
		/* m.essage("first %s", (gchar *) walk->data); */
		bond_api_addloadby(bond, widgetname, (gchar *) walk->data);
		}
	/* m.essage("now painting. %s", widgetname); */
	bobj = bond_getbyname(bond, widgetname);
	/* If hasn't been spawned, spawn it. */
	if (!bobj)
		{
		/* m.essage("spawning bond object now"); */ 
		bond_api_open(bond, widgetname);
		}
	else
		{
		/* refresh the object with the new loadby */
		bond_api_refresh(bond, widgetname);
		/* make it visible if its hidden */
		if (ui_isvisible(bond->uixml, bobj->widget) == FALSE)
			bond_show(bobj, FALSE);
		}

	return 0;
	}

/**
 * bond_api:
 * @call: enum of the api call to make, defined in bondxml.h.
 * 
 * bond_api will run the api function associated with the call made. For example
 * if the call was "open", bond_api will run
 * bond_api_open(bond,va_arg(ap,gchar*));
 *
 * Returns: Non-zero on error.
 */
gint
bond_api(Bond * bond, ApiCall call, ...)
	{
	gchar *name, *parentname, *sql;
	GList *list;
	gint retval = 0;
	va_list ap;

	g_assert(bond);
	va_start(ap, call);

	switch (call)
		{
	case ApiNone:
		break;
	case ApiOpen:
		bond_api_open(bond, va_arg(ap, gchar *));
		break;
	case ApiClose:
		bond_api_close(bond, va_arg(ap, gchar *));
		break;
	case ApiAdd:
		bond_api_add(bond, va_arg(ap, gchar *));
		break;
	case ApiNew:
		bond_api_new(bond, va_arg(ap, gchar *));
		break;
	case ApiDelete:
		bond_api_delete(bond, va_arg(ap, gchar *));
		break;
	case ApiLoadAll:
		bond_api_loadall(bond, va_arg(ap, gchar *));
		break;
	case ApiRefresh:
		bond_api_refresh(bond, va_arg(ap, gchar *));
		break;
	case ApiClear:
		bond_api_clear(bond, va_arg(ap, gchar *));
		break;
	case ApiFlush:
		bond_saveall(bond);
		db_toliet_flushall();
		break;
	case ApiNext:
		bond_api_next(bond, va_arg(ap, gchar *));
		break;
	case ApiPrev:
		bond_api_prev(bond, va_arg(ap, gchar *));
		break;
	case ApiFirst:
		bond_api_first(bond, va_arg(ap, gchar *));
		break;
	case ApiLast:
		bond_api_last(bond, va_arg(ap, gchar *));
		break;
	case ApiSave:
		bond_api_save(bond, va_arg(ap, gchar *));
		break;
	case ApiSQL:
		name = va_arg(ap, gchar *);
		parentname = va_arg(ap, gchar *);
		sql = va_arg(ap, gchar *);
		bond_api_sql(bond, name, parentname, sql);
		break;
	case ApiLoadBy:
		name = va_arg(ap, gchar *);
		list = va_arg(ap, void *);

		bond_api_loadby(bond, name, list);
		break;
	case ApiQuit:
		ui_mainquit(bond->uixml);
		break;
	case ApiMoveTo:
		break;
	case ApiAbortSave:
		bond_api_abortsave(bond, va_arg(ap, gchar *));
	case ApiReport:

		break;
	default:
		errormsg("Unknown API call %d", call);
		break;
		}

	va_end(ap);
	return retval;
	}

/**
 * bond_api_byquery:
 * @formname: Name of a form (ie window) to show.  It relate this to a table name.
 * @sqlquery: SQL query to run and result will be sent to @formname.
 *
 * Excute sql query @sqlquery, and create a database object to these results.
 * These results will be displayed to form @formname.  For example if @formname
 * allows you to cycle though records only the record returned from @sqlquery
 * will be shown.
 *
 * Returns: non-zero on error. return value will be 1 if @formname needs saving first
 * by bond_api_save().
 */

/**
 * bond_api_byfield: 
 * @formname: Name of a window to display, which refers to a tablename
 * @field: A field that can be associated with @formname.
 * @value: The value you which to filter on, so that it will show all fields which = @field.
 * 
 * Shows a form, but filters it on a specific field been equal to a certain value.
 * It is like calling a SELECT * FROM tablename WHERE field = value.  This
 * api call replaces bond_api_loadbyvalue(), any old code from 0.2 or lower
 * needs to replace there API calls to that function to this function.
 * 
 * 
 * Returns: non-zero on error.
 */

/* Similar to db lower level function. */

/**
 * bond_api_search:
 * @inputlist: A widget that contains the components inwhich to build the search requirements
 * @outputlist: Widget to show output, generally a GtkCList.
 * 
 * Build a list to search from... umdedoeddaaa. 
 */

/**
 * bond_api_searchbyobject:
 * 
 * Alteration of the bond_api_search() to show the results of a search, generated independently and
 * stored in object obj, in a clist, named outputlist.  This is to allow for searches which don't
 * relate perfectly to fieldx = blah and fieldy = otherblah - like fieldx < blah and fieldx > otherblah
 * sort of thing.  But basically copied from the above function anyway.
 *
 * Returns: non-zero on error.
 */

/**
 * bond_api_showbysearch:
 * @searchresult: A list (normally clist) of search results generated by a bond_api_search()
 * @showform: Name of a form/frame etc, to show from the position in list.
 * @position: Position in list to show in the other record.
 * 
 * @inputlist Referes to a roof which has an objectlist.  It could a clist in a form or the form itself.
 * @searchresult refers to the form to load and show the results, which could be the same name as the form
 * it derived from.  The @position referes to the row that was selected.  
 * 
 * This is a useful function to show the a form or other widget based on what
 * you selected from a search. Ie bring up a persons or sales details after you
 * clicked on it.
 * 
 * Returns: Non zero on error.
 */

/**
 * bond_showformbyform:
 * @formname: New form to load.
 * @oldform: The old form which your loading this object from.
 *
 * Find a list or reference to the subform classname which is in the
 * main class name.  Get the destination objectid, and the localid,
 * use that information to retrieve either a single object or an
 * objectlist.
 * 
 * In regard to subforms, they can be complex and include information
 * like a list and a single object.  . .. wait I'll think about this
 * and be back later. Ok I thought about it.  If their is a clist in
 * the widget set as well as an objectlist present then drop that
 * ginto that clist.  Check first the list picked doesn't have a field
 * referencing that list.
 *
 * Returns: Non-zero on error.
 */

/**
 * bond_api_showrecord:
 * @formname: Name of a form.
 * @recordpos: Position offset.
 * 
 * Move to the record position specified by @recordpos.
 *
 * Returns: non-zero on error.
 */

/**
 * bond_flushall:
 * 
 * Flush any writes to the database that are due.
 * This shouldn't normally need to be called, but it can be useful in some
 * circumstances.  Any changes that havn't been written are forced to the
 * backend database.
 */

/**
 * bond_showformbyformwithquery:
 * @formname: Formname to show
 * @oldform: UiForm you are showing this from (the parent)
 * @sqlquery: A SQL query to excute which will make up the contents of @formname
 * 
 * Excute the query @sqlquery and display the results in @formname.  It will be treated
 * as a child form of @oldform. For example if @formname
 * allows you to cycle though records only the record returned from @sqlquery
 * will be shown.
 * Note that the only subset of SQL should be used.
 * 1. Only a single SELECT query is allowed.
 * 2. FROM clause should contain only ONE table which is to be displayed as a new form.
 * 3. The @forname must contain a reference field to the @oldform, and it has to be
 *    appeared in SELECT 
 * 4. Due to the SQL parsing problem, only simple WHERE clause conditions such as
 *    table1.field1  = table2.filed2 AND table3.field3 = 1 (a constant value)
 * 5. Do not use comparison operator such as >, <. However, you are OK to use >=, <=
 *    In fact, using >=, <=, is a wicked way to get around the problem.
 * 6. Do not use functions such as date('today'). In this case which will return a
 *    simple value, use db_sql() to get constant value, then use the value with this
 *    function.
 * 7. You better not try using ORDER BY clause here.
 *
 * If you want to improve this look at dbsqlparse.c and dbbirth.c, etc
 * The trouble starts from db_birth_createfromsql() in this code.
 *
 * Returns: non-zero on error. return value will be 1 if @formname needs saving$ * by bond_api_save().
 */
