/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>

#include "db.h"
#include "dbfilter.h"

#include "bondfilter.h"
#include "watchtower.h"
#include "bondrelation.h"
#include "bond.h"
#include "ui.h"

static WatchTower *
bondfilter_buildwatchtower(BondObject * bobj, OperationsNodeFilter * oprfilter,
                           DbFilterAppend * filter)
	{
	BondObject *redrawbobj;
	WatchTower *watchtower;

	g_assert(bobj);
	g_assert(oprfilter);

	watchtower = watchtower_create(bobj, bondfilter_updatefilter);
	watchtower->ptr[0] = filter;
	watchtower->ptr[1] = oprfilter;
	redrawbobj = bobj;
	if (bobj->widget == NULL)
		if (bobj->parent && bobj->parent->widget)
			redrawbobj = bobj->parent;
	watchtower->ptr[2] = redrawbobj;

	watchtower->tablename = oprfilter->value_table;
	watchtower->fieldname = oprfilter->value_field;
	watchtower->dbobjectsrc = oprfilter->value_dbobjectsrc;
	if (oprfilter->value_widget)
		watchtower->widget =
		    ui_getwidget(bobj->bond->uixml, oprfilter->value_widget);

	return watchtower;
	}

static GList *
bondfilter_applyfilters_getlist(BondObject * bobj, OperationsNodeDb * dbnode)
	{
	OperationsNodeFilter *oprfilter;
	WatchTower *watchtower;
	DbFilterAppend *filter;
	GList *retlist = NULL, *walk, *tmplist;

	for (walk = g_list_first(dbnode->filters); walk != NULL; walk = walk->next)
		{
		oprfilter = walk->data;
		filter = mem_calloc(sizeof(DbFilterAppend));
		if (oprfilter->table)
			filter->tablename = mem_strdup(oprfilter->table);
		filter->fieldname = mem_strdup(oprfilter->field);
		filter->fieldpos =
		    db_field_getpos(bobj->obj, oprfilter->field, oprfilter->table);
		filter->operation = oprfilter->opr;
		if (oprfilter->value_constant)
			filter->value = mem_strdup(oprfilter->value_constant);
		retlist = g_list_append(retlist, filter);

		if (oprfilter->value_field || oprfilter->value_widget)
			{
			watchtower = bondfilter_buildwatchtower(bobj, oprfilter, filter);
			tmplist = watchtower_register(watchtower);
			if (tmplist)
				g_list_free(tmplist);
			}
		}

	return retlist;
	}

/**
 * bondfilter_updatefilter:
 * @ptr1:
 * @ptr2:
 * 
 * Run by watch process. 
 * 
 * Returns: Nothing at all.
 */
void
bondfilter_updatefilter(void *ptr1, void *ptr2)
	{
	WatchTower *watchtower;
	BondObject *watched, *redraw;
	OperationsNodeFilter *oprfilter;
	Object *objchanged;
	gchar *value = NULL;
	DbFilterAppend *filter;

	/* initialise everything */
	watchtower = ptr1;
	watched = ptr2;
	g_assert(watchtower->watching);
	g_assert(watched);
	debuglog(90, "%s changed trigging %s to update.", watched->widgetname,
	         watchtower->watching->widgetname);
	/* watched is the object that the change occurs. watchtower->watching is the
	   one with the object with filter */
	if (watchtower->watching->obj == NULL)
		{
		errormsg("No database object set with filter in it.");
		return ;
		}
	filter = watchtower->ptr[0];
	oprfilter = watchtower->ptr[1];
	redraw = watchtower->ptr[2];
	/* clean old values */
	g_assert(filter);
	if (filter->value)
		{
		mem_free(filter->value);
		filter->value = NULL;
		}
	/* get new value */
	value = NULL;
	if (oprfilter->value_dbobjectsrc)
		{
		objchanged =
		    bondrelation_findobjfromobjsrc(watched->bond,
		                                   oprfilter->value_dbobjectsrc);
		if (objchanged == NULL)
			{
			warningmsg("Can't run filter %s due to %s not been loaded yet.",
			           watchtower->watching->widgetname,
			           oprfilter->value_dbobjectsrc);
			return ;
			}
		if (oprfilter->value_field)
			db_getvalue(objchanged, oprfilter->value_field,
			            oprfilter->value_table, &value);
		}
	else if (oprfilter->value_constant)
		value = oprfilter->value_constant;
	else if (oprfilter->value_widget)
		{
		if (watched == NULL)
			watched = bond_getbyname(watched->bond, oprfilter->value_widget);
		bond_save_getui(watched, &value);
		}

	/* Update filter, rerun, and update object now */
	if (value)
		{
		/* m.essage("updating filter %s.%s to be eqaul to %s on
		   %s",filter->tablename,filter->fieldname,value,watchtower->watching->obj->basetable); 
		 */
		filter->value = mem_strdup(value);
		watchtower->functionhasbeenrun = TRUE;
		/* if the widget is already drawn redraw it, else ignore and leave it
		   for itself to draw */
		if (redraw->ready == TRUE)
			{
			/* update the filter in the orginal object */
			db_filter_apply(watchtower->watching->obj);
			/* db_cache_debug(watchtower->watching->obj); */
			/* m.essage("Repainting %s",redraw->widgetname); */
			bond_paint(redraw, FALSE, TRUE);
			}
		/* else warningmsg("Widget %s not ready to be redrawn. Assuming yet to
		   be drawn",redraw->widgetname); */
		}
	watchtower->ready = TRUE;
	}

/**
 * bondfilter_applyfilters:
 * @bobj: Bond object to act upon
 *
 * Load any filters we need to. Filters will be updated later with there current
 *	values. 
 *
 * Returns: Non-zero on error.
 */
gint
bondfilter_applyfilters(BondObject * bobj)
	{
	OperationsNodeDb *dbnode;
	GList *filterlist;

	if (bobj->obj == NULL)
		return -1;
	if (bobj->oprnode->dbobj->filters == NULL)
		return 0;
	dbnode = bobj->oprnode->dbobj;
	/* Filter the object in bonddb */
	filterlist = bondfilter_applyfilters_getlist(bobj, dbnode);
	if (filterlist == NULL)
		{
		errormsg("Expected a GList* here");
		return -2;
		}

	if (bobj->obj->filterapplied)
		db_filter_freeapplied(bobj->obj);
	db_filter_load(bobj->obj, filterlist);
	db_filter_createpos(bobj->obj);

	return 0;
	}
