/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version. 
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <gmodule.h>

#include "bc.h"

#include "bond.h"
#include "bondxml.h"
#include "bondapi.h"
#include "bondmain.h"
#include "bondtrigger.h"
#include "bondargument.h"

static gint
bondtrigger_runapi(BondObject * bobj, OperationsNodeTrigger * trigger)
	{
	gint retval = 0;

	g_assert(bobj);
	g_assert(trigger);

	switch (trigger->apicall)
		{
	case ApiNone:
		retval = -1;
		break;
	case ApiOpen:
	case ApiClose:
	case ApiAdd:
	case ApiNew:
	case ApiDelete:
	case ApiLoadAll:
	case ApiRefresh:
	case ApiNext:
	case ApiPrev:
	case ApiFirst:
	case ApiLast:
	case ApiClear:
	case ApiSave:
		if (trigger->apiwidgetname)
			{
			retval = bond_api(bobj->bond, trigger->apicall,
				trigger->apiwidgetname);
			break;
			}
		if (bobj->dbobjectsrc)
			{
			debuglog(30,
			         "Widgetname should be specified on %s. "
				 "Guessing now", bobj->widgetname);
			retval =
			    bond_api(bobj->bond, trigger->apicall,
			             bobj->dbobjectsrc->widgetname);
			break;
			}
		errormsg("No widgetname is specified in api trigger in %s",
		         bobj->widgetname);
		retval = -1;
		break;
	case ApiSQL:
		retval =
		    bond_api(bobj->bond, ApiSQL, trigger->apiwidgetname,
		             bobj->widgetname, trigger->sql);
		break;
	case ApiLoadBy:
		retval =
		    bond_api(bobj->bond, ApiLoadBy, trigger->apiwidgetname,
		             trigger->loadby);
		break;
	case ApiFlush:
	case ApiQuit:
		retval = bond_api(bobj->bond, trigger->apicall);
		break;
	case ApiMoveTo:
		break;
	case ApiReport:
		break;
	default:
		errormsg("Unrecognised API call.");
		retval = -2;
		break;
		}
	return retval;
	}

static void *
bondtrigger_getargument(BondObject *bobj, OperationsNodeTriggerArgument *arg)
	{
	BondObject *retbobj;
	gchar *retstr;
	UiWidget *widget;
	
	if (arg->widgetname)
		{
		retbobj = bond_getbyname(bobj->bond, arg->widgetname);
		bond_save_getui(retbobj, &retstr);
		return (void*)retstr;
		}
	if (arg->constant)
		return (void*)arg->constant;
	if (arg->dbobjectsrc)
		{
		return (void*)bond_getbyname(bobj->bond, arg->dbobjectsrc);
		}
	if (arg->table)
		{
		warningmsg("Unforuntly this is not coded yet. Remind me!");
		}
	return NULL;
	}	
	
static void *
bondtrigger_getfunc(Bond * bond, gchar * funcname)
	{
	void *ptrtofunc = NULL;

	g_assert(bond->symbols);

	if (!g_module_symbol(bond->symbols, funcname, &ptrtofunc))
		errormsg("Unable to find function %s specified in the xml "
			"file.",funcname);
	return ptrtofunc;
	}

static gint
bondtrigger_runfunction(BondObject * bobj, OperationsNodeTrigger * trigger)
	{
	void *value;
	GList *arglist = NULL, *walk;
	typedef gint(*function) (BondObject *, GList *);
	function func;
	OperationsNodeTriggerArgument *arg;

	if (!trigger->ptr2func)
		{
		trigger->ptr2func =
		    bondtrigger_getfunc(bobj->bond, trigger->functionname);
		if (!trigger->ptr2func)
			return -1;
		}
	for (walk=g_list_first(trigger->argumentlist);walk!=NULL;
			walk=walk->next)
		{
		arg = walk->data;
		value = bondtrigger_getargument(bobj, arg);
		arglist = g_list_prepend(arglist, value);		
		}
	
	func = trigger->ptr2func;
	return func(bobj, arglist);
	}

static gint
bondtrigger_runvaluelink(BondObject * bobj, OperationsNodeTrigger * trigger)
	{
	errormsg("NOT WRITTEN");
	return 0;
	}

static gint
bondtrigger_runshell(BondObject * bobj, OperationsNodeTrigger * trigger)
	{
	message("Calling shell script");
	system(trigger->shell);
	return 0;
	}

static gint
bondtrigger_runpapyrus(BondObject *bobj, OperationsNodeTrigger * trigger)
	{
	message("Calling papyrus gui");
	system("papyrus_gui");
	}

/**
 * bondtrigger_runreport
 *
 * @trigger: An operations node trigger. For this function, @trigger contains
 * values parsed from the XML file, such as the report file name and type and
 * arguments entered. 
 *
 * Will process report information to create a command string and execute the
 * command. Values such as the directory where /papyrus/ and the directory where
 * the report files are located are found in the program's configuration
 * file. Values such as the report file name and type are retrieved from the
 * program's xml file.
 *
 * Returns: If an error occurred -1.
 */

static gint
bondtrigger_runreport(BondObject * bobj, OperationsNodeTrigger * trigger)
	{
	BondObject *argbobj;	/* A bond object to get arg widget values.. */
	gchar *papyrus_str;	/* string with papyrus path and name */
	gchar *report_str;	/* path of report file path and name */
	gchar *output_str;	/* path of output file path and name */
	gchar *arg_str = "";	/* argument list string */
	gchar *argvalue;
	gchar *command_str;	/* the complete command string */
	GList *walk;		/* walk through argument list */
	OperationsNodeTriggerArgument *argument;	/* an argument in the 
				 * argument list */
	gint retval = 0;

	/* create papyrus full path string */
	papyrus_str = bcconfig_getvalue("papyrusdir");
	if (papyrus_str)
		{
		if (*(papyrus_str + strlen(papyrus_str) - 1) != '/')
			papyrus_str = g_strconcat(papyrus_str, "/papyrus", 
					NULL);
		}
	else
		papyrus_str = "/usr/local/bin/papyrus";

	/* create report file full path string */
	report_str = bcconfig_getvalue("reportdir");
	if (report_str)
		{
		if (*(report_str + strlen(report_str) - 1) != '/')
			report_str = g_strconcat(report_str, "/", NULL);
		}
	else
		report_str = "./";

	/* create output file full path string */
	output_str = bcconfig_getvalue("outputdir");
	if (output_str)
		{
		if (*(output_str + strlen(output_str) - 1) != '/')
			output_str = g_strconcat("-o ", output_str, "/", NULL);
		else
			output_str = g_strconcat("-o ", output_str, NULL);
		}
	else												 /* don't output file */
		output_str = NULL;

	/* walk through GList, concatenate arg's together into string */
	for (walk = g_list_first(trigger->argumentlist); walk != NULL;
	        walk = walk->next)
		{
		/* add prefix "-a " for each argument and append to argument 
		 * string */
		argument = walk->data;	 /* an argument in the argument list */
		if (argument->name)
			{
			argbobj = bond_getbyname(bobj->bond, 
					argument->widgetname);
			bond_save_getui(argbobj, &argvalue);
			if (strcmp(argument->name, "report_file") == 0)
				{
				trigger->reportfile = argvalue;
				}
			else if (argvalue && (strlen(argvalue) > 0))
				{
				debugmsg("%s=%s", argument->name, argvalue);
				arg_str = g_strconcat(arg_str, "-a ", 
					argument->name, "=",argvalue, " ", 
					NULL);
				}
			}
		}

	/* add report file name to end of report and output string */
	if (trigger->reportfile)
		{
		report_str = g_strconcat(report_str, trigger->reportfile, NULL);
		if (output_str)
			output_str = g_strconcat(output_str, 
				trigger->reportfile,".", 
				trigger->reportfiletype, NULL);
		}
	else
		{
		errormsg("Could not find report file in the trigger");
		return -1;
		}

	if (output_str)
		{
		if (trigger->reportpreview)
			/* if report preview bit TRUE add "-v" */
			command_str = g_strjoin(" ", papyrus_str, arg_str, 
					report_str, "-t", 
					trigger->reportfiletype, "-v",
			                output_str, NULL);
		else
			command_str = g_strjoin(" ", papyrus_str, arg_str, 
					report_str, "-t", 
					trigger->reportfiletype, output_str,
			                NULL);
		}
	else
		{
		if (trigger->reportpreview)
			command_str = g_strjoin(" ", papyrus_str, arg_str, 
					report_str, "-t", 
					trigger->reportfiletype, "-v", NULL);
		else
			command_str = g_strjoin(" ", papyrus_str, arg_str, 
					report_str, "-t", 
					trigger->reportfiletype, NULL);
		}

	/* For our reports, we have a DTD so lets turn it on, eh? */
	command_str = g_strjoin(" ", command_str, "--validate", NULL);

	retval = system(command_str);	/* execute the command */

	/* free up all strings in memory */
	/* g_free(arg_str); */
	/* g_free(report_str); */
	/* g_free(papyrus_str); */
	/* g_free(command_str); */
	debugmsg("This is the command string to be executed: \n %s\n", 
			command_str);
	return retval;
	}

gint
bondtrigger_action(BondObject * bobj, OperationsNodeTrigger * trigger)
	{
	gchar *message;
	gint retval = 0;

	g_assert(bobj);
	g_assert(trigger);

	debuglog(50, "Trigger in widget %s has been actived.", 
			bobj->widgetname);
	/* normally only one trigger type action will exist but i added 
	 * support for all of them for those special cases. */
	if (trigger->apicall > 0)
		{
		debuglog(30, "API call called");
		/* If you need to supply a confirmation box then do so. */
		if (trigger->confirmaction == TRUE)
			{
			if (trigger->apicall == ApiDelete)
				message = mem_strdup("Are you sure you wish "
					"to delete this record?");
			else
			if (trigger->apicall == ApiQuit)
				message = mem_strdup("Are you sure you want "
					"to quit?");
			else
				message = mem_strdup("Are you sure?");
			if (ui_display_message(bobj->bond->uixml,message,
					"Question",2)==0)
				;
			else
				{
				mem_free(message);
				return -1;
				}
			mem_free(message);
			}
		retval = bondtrigger_runapi(bobj, trigger);
		}
	if (trigger->functionname)
		retval = bondtrigger_runfunction(bobj, trigger);
	if (trigger->valuelink)
		retval = bondtrigger_runvaluelink(bobj, trigger);
	if (trigger->reportfile)
		retval = bondtrigger_runreport(bobj, trigger);
	if (trigger->shell)
		retval = bondtrigger_runshell(bobj, trigger);
	if (trigger->runpapyrus)
		retval = bondtrigger_runpapyrus(bobj, trigger);

	return retval;
	}

gint
bondtrigger_init(Bond * bond)
	{
	if (!g_module_supported())
		{
		errormsg("The application needs g_module support before it "
				"can have signals");
		return -1;
		}

	bond->symbols = (GModule *) g_module_open(NULL, 0);
	return 0;
	}

void
testbondfunc(BondObject * bobj, GList * args)
	{
	attention("I AM TIGER!");

	}

gint
bondtrigger_checkevent(BondObject *bobj, TriggerEvent event)
	{
	GList *walk;
	OperationsNodeTrigger *trigger;
	
	for (walk=g_list_first(bobj->oprnode->triggerlist);
	     walk!=NULL;walk=walk->next)
	     {
	     trigger = walk->data;
	     if (trigger->event == event)
		  {
		  bondtrigger_action(bobj,trigger);
		  return 0;
		  }
	     }
	return 1;
	}
