/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include <glib.h>
#include <gtk/gtk.h>
#include <libxml/parser.h>

#include "bc.h"
#include "parse.h"
#include "xml_funcs.h"
#include "xml_display.h"

#include "uidropdown.h"
#include "uilist.h"
#include "uigeneral.h"
#include "uibox.h"
#include "uispinbutton.h"
#include "uitoggle.h"
#include "uiprogressbar.h"
#include "uiimage.h"
#include "uinotebook.h"
#include "uiframe.h"
#include "uiseparator.h"
#include "uibutton.h"
#include "uimenu.h"
#include "uiradiobutton.h"
#include "uicheckbutton.h"
#include "uitoolbar.h"
#include "uitextview.h"

static gchar *
uixml_generateshortname(xmlNode * widgetnode, gchar * widgetname, UiXML * bond)
	{
	/* xmlNode *child; */
	gchar *retstr = NULL;

	retstr = xmlGetProp(widgetnode, "name");
	if (!retstr)
		{
		retstr =
		    g_strdup_printf("%s%d", widgetnode->name, bond->widgetcounter++);
		xmlNewProp(widgetnode, "name", retstr);
		}
	return retstr;
	}

static void
uixml_generatepath(xmlNode * widgetnode, UiXML * bond, gchar ** longname,
                   gchar ** shortname)
	{
	xmlNodePtr node;
	gchar *content = NULL;

	*shortname = uixml_generateshortname(widgetnode, *shortname, bond);

	node = widgetnode;
	while (!content && node)
		{
		content = xmlGetProp(node, "longname");
		node = node->parent;
		}
	if (content)
		*longname = g_strdup_printf("%s/%s", content, *shortname);
	else
		*longname = g_strdup(*shortname);
	xmlNewProp(widgetnode, "longname", *longname);

	debuglog(100, "Shortname %s, Longname: %s", *shortname,
	         xmlGetProp(widgetnode, "longname"));
	}

static gint
uixml_parsewidget_attribute(xmlNode * node, UiWidget * widget, gchar ** name,
                            UiXML * uixml)
	{
	gint minheight = -1, minwidth = -1;
	gchar *content;
	gboolean isscrolledwindow = TRUE;

	g_assert(widget);

	*name = NULL;

	content = xmlGetProp(node, "name");
	if (content)
		*name = content;

	content = xmlGetProp(node, "minwidth");
	if (content)
		minwidth = atoi(content);

	content = xmlGetProp(node, "minheight");
	if (content)
		minheight = atoi(content);
	
	content = xmlGetProp(node, "scrolledwindow");
	if (content && IsFalse(content))
		isscrolledwindow = FALSE;

	if (minheight > 0 || minwidth > 0)
		{
		if (uigtk_gettype(uixml, widget) == UiGrid &&
				isscrolledwindow == TRUE)
			{
			minheight -= 5;
			minwidth -= 3;
			}
		else if (uigtk_gettype(uixml, widget) != UiGridColumn)
		     	uigtk_setwidgetsize(uixml, widget, minwidth, minheight);
		}

	content = xmlGetProp(node, "grabfocus");
	if (content && IsTrue(content))
	{
		gtk_widget_grab_focus(widget);
		attention("FOCUS GRAB %s",name);
	}
	
	content = xmlGetProp(node, "defaultfocus");
	if (content && IsTrue(content))
		gtk_widget_grab_default(widget);
			
	return 0;
	}

static void
uixml_parse_project(UiXML * uixml, xmlNodePtr parentnode)
	{
	gchar *content;

	g_assert(uixml);

	content = xmlGetProp(parentnode, "defaultvisible");
	if (content && IsTrue(content))
		uixml->defaultvisible = TRUE;
	else
		uixml->defaultvisible = FALSE;

	content = xmlGetProp(parentnode, "defaultentrywidget");
	if (content)
		uixml->defaultentrywidth = atoi(content);
	}

int
uixml_node(xmlNode * node, gchar * name)
	{
	if (IsEqual(node->name, name))
		return 1;
	return 0;
	}

int
uixml_parseparent(UiXML * bondxml)
	{
	xmlNodePtr node;

	node = xmlDocGetRootElement(bondxml->doc);
	xml_display_doc(bondxml->doc);
	node = node->children;

	uixml_parse(node, 0, NULL, bondxml);
	return 0;
	}

UiXML *
uixml_init(gchar * filename)
	{
	gchar *dtd_value;
	gchar *xi_value;
	xmlDocPtr doc;
	xmlParserCtxtPtr dtd_ctxt;	  /* the parser context */

	UiXML *bondxml;

	/* create a parser context */
	dtd_ctxt = xmlNewParserCtxt();

	if (!dtd_ctxt)
		{
		fprintf(stderr, "Failed to allocate parser context\n");
		return NULL;
		}

	/* Get configuration options */
	xi_value = bcconfig_getvalue("xinclude_processing");
	g_message("xinclude_processing = %s", xi_value);
	dtd_value = bcconfig_getvalue("dtd_validation");

	/* Check if the file needs to be validated with dtd */
	/* parse the file, activating the DTD validation option */
	if (dtd_value && IsTrue(dtd_value))
		{
		/* Check if XInclude processing should be enabled */
		if (xi_value && IsTrue(xi_value))
			doc = xmlCtxtReadFile(dtd_ctxt,
			                      filename,
			                      NULL,
			                      XML_PARSE_DTDVALID | XML_PARSE_XINCLUDE);
		else
			doc = xmlCtxtReadFile(dtd_ctxt, filename, NULL, XML_PARSE_DTDVALID);
		}
	else if (xi_value && IsTrue(xi_value))
		doc = xmlCtxtReadFile(dtd_ctxt, filename, NULL, XML_PARSE_XINCLUDE);
	else /* otherwise, parse without validation */
		doc = xmlCtxtReadFile(dtd_ctxt, filename, NULL, 0);

	/* check if parsing suceeded */
	if (!doc)
		{
		errormsg("Error parsing xmlfile %s.\n", filename);
		return NULL;
		}
	else
		{
		/* check if validation suceeded */
		if (dtd_ctxt->valid == 0)
			{
			errormsg("Failed to validate %s\n", filename);
			return NULL;
			}
		}
	debugmsg("Parsing the document");
	bondxml = mem_calloc(sizeof(UiXML));
	bondxml->doc = doc;
	bondxml->defaultvisible = TRUE;
	bondxml->defaultentrywidth = -1;
	/* Set up hash tables which contain widgets and lookups */
	bondxml->widgets =
	    g_hash_table_new_full(&g_direct_hash, &g_direct_equal, NULL, NULL);
	bondxml->shortname =
	    g_hash_table_new_full(&g_str_hash, &g_str_equal, &g_free, NULL);
	bondxml->longname =
	    g_hash_table_new_full(&g_str_hash, &g_str_equal, &g_free, NULL);
	bondxml->garbage = NULL;
	/* This should free up the context (but not the dtd_ctxt->doc) */
	xmlFreeParserCtxt(dtd_ctxt);
	debugmsg("XML context freed (except the XML tree)");

	return bondxml;
	}

UiWidget *
uixml_parse(xmlNode * node, ParseState state, UiWidget * parentwidget,
            UiXML * bondxml)
	{
	UiWidget *widget = NULL, *tmpwidget = NULL, *scrolledwindow = NULL;
	gchar *tmpstr1 = NULL, *tmpstr2 = NULL;
	guint poscounter = 0;
	gboolean visible;
	gchar *name = NULL;

	for (node = node->children; node; node = node->next)
		{
		/* The name of the current node */
		name = (char *)node->name;
		switch (state)
			{
		case ParseStart:
			if (IsEqual(name, "project"))
				{
				uixml_parse_project(bondxml, node);
				break;
				}
			if (IsEqual(name, "window"))
				{
				widget = uigtk_window_create(bondxml, node, 
					parentwidget, &visible);
				uixml_parsename(node, widget, bondxml);
				tmpwidget =
				    uixml_parse(node, ParseGeneral, widget, bondxml);
				if (tmpwidget)
					{
					gtk_container_add(GTK_CONTAINER(widget),
					                  tmpwidget);
					}
				if (visible == TRUE)
					uigtk_showwidget(bondxml, widget);
				else
					uigtk_hidewidget(bondxml, widget);
				break;
				}
			break;
		case ParseToolBar:
			if (IsEqual(name, "toolbutton"))
				{
				widget = uigtk_toolbar_toolbutton(node, parentwidget);
				uixml_parsename(node, widget, bondxml);
				break;
				}
			if (IsEqual(name, "toolbarspace"))
				{
				uigtk_toolbar_add_space(parentwidget);
				break;
				}
		case ParseHandleBox:
		case ParseGeneral:
			if (IsEqual(name, "label"))
				{
				widget = uigtk_label_create(node, parentwidget);
				uixml_parsename(node, widget, bondxml);
				break;
				}
			if (IsEqual(name, "entry"))
				{
				widget =
				    uigtk_entry_create(bondxml, node, parentwidget);
				uixml_parsename(node, widget, bondxml);
				break;
				}
			if (IsEqual(name, "dropdown"))
				{
				widget = uigtk_dropdown_create(node, parentwidget);
				uixml_parsename(node, widget, bondxml);
				break;
				}
			if (IsEqual(name, "box"))
				{
				widget = uigtk_box_create(node, parentwidget);
				if (widget->parent)
					errormsg("Widgetparent isn't null!");
				uixml_parsename(node, widget, bondxml);
				uixml_parse(node, ParseBox, widget, bondxml);
				break;
				}
			if (IsEqual(name, "vbox"))
				{
				widget = uigtk_vhbox_create(node, parentwidget, 1);
				uixml_parsename(node, widget, bondxml);
				uixml_parse(node, ParseVHBox, widget, bondxml);
				break;
				}
			if (IsEqual(name, "hbox"))
				{
				widget = uigtk_vhbox_create(node, parentwidget, 0);
				uixml_parsename(node, widget, bondxml);
				uixml_parse(node, ParseVHBox, widget, bondxml);
				break;
				}
			if (IsEqual(name, "button"))
				{
				widget = uigtk_button_create(node, parentwidget);
				uixml_parsename(node, widget, bondxml);
				break;
				}
			if (IsEqual(name, "list"))
				{
				widget =
				    uigtk_list_create(node, parentwidget,
				                      &scrolledwindow, bondxml);
				if (scrolledwindow)
					uixml_parsewidget_attribute(node, scrolledwindow,
					                            &tmpstr1, bondxml);
				uixml_parsename(node, widget, bondxml);
				break;
				}
			if (IsEqual(name, "spin"))
				{
				widget = uigtk_spinbutton_create(node, parentwidget);
				uixml_parsename(node, widget, bondxml);
				break;
				}
			if (IsEqual(name, "toggle"))
				{
				widget = uigtk_toggle_button_create(node, parentwidget);
				uixml_parsename(node, widget, bondxml);
				break;
				}
			if (IsEqual(name, "check"))
				{
				widget = uigtk_check_button_create(node, parentwidget);
				uixml_parsename(node, widget, bondxml);
				break;
				}
			if (IsEqual(name, "scale"))
				{
				widget = uigtk_scale_create(node, parentwidget);
				uixml_parsename(node, widget, bondxml);
				break;
				}
			if (IsEqual(name, "separator"))
				{
				widget = uigtk_separator_create(node, parentwidget);
				uixml_parsename(node, widget, bondxml);
				break;
				}
			if (IsEqual(name, "frame"))
				{
				widget = uigtk_frame_create(node, parentwidget);
				uixml_parsename(node, widget, bondxml);
				tmpwidget =
				    uixml_parse(node, ParseGeneral, widget, bondxml);
				if (tmpwidget)
					gtk_container_add(GTK_CONTAINER(widget),
					                  tmpwidget);
				break;
				}
			if (IsEqual(name, "scrollbar"))
				{
				widget = uigtk_scroll_bar_create(node, parentwidget);
				uixml_parsename(node, widget, bondxml);
				break;
				}
			if (IsEqual(name, "image"))
				{
				widget = uigtk_image_create(node, parentwidget);
				uixml_parsename(node, widget, bondxml);
				break;
				}
			if (IsEqual(name, "progressbar"))
				{
				widget = uigtk_progress_bar_create(node, parentwidget);
				uixml_parsename(node, widget, bondxml);
				break;
				}
			if (IsEqual(name, "statusbar"))
				{
				widget = uigtk_status_bar_create(node, parentwidget);
				if (!widget)
					break;
				uixml_parsename(node, widget, bondxml);
				break;
				}
			if (IsEqual(name, "radio"))
				{
				widget = uigtk_radio_button_create(node, parentwidget);
				uixml_parsename(node, widget, bondxml);
				uixml_parse(node, ParseRadio, widget, bondxml);
				break;
				}
			if (IsEqual(name, "notebook"))
				{
				widget = uigtk_notebook_create(node, parentwidget);
				uixml_parsename(node, widget, bondxml);
				uixml_parse(node, ParseNotebook, widget, bondxml);
				break;
				}
			if (IsEqual(name, "menubar"))
				{
				widget = uigtk_menubar_create(node, parentwidget);
				uixml_parsename(node, widget, bondxml);
				uixml_parse(node, ParseMenuBar, widget, bondxml);
				break;
				}
			if (IsEqual(name, "toolbar"))
				{
				widget = uigtk_toolbar_create(node, parentwidget);
				uixml_parsename(node, widget, bondxml);
				uixml_parse(node, ParseToolBar, widget, bondxml);
				break;
				}
			if (IsEqual(name, "handlebox"))
				{
				widget =
				    uigtk_handlebox_create(bondxml, node, parentwidget);
				uixml_parsename(node, widget, bondxml);
				uixml_parse(node, ParseHandleBox, widget, bondxml);
				break;
				}
			if (IsEqual(name, "textview"))
				{
				widget =
				    uigtk_textview_create(node, parentwidget,
				                          &scrolledwindow);
				if (scrolledwindow)
					uixml_parsewidget_attribute(node, scrolledwindow,
					                            &tmpstr1, bondxml);
				uixml_parsename(node, widget, bondxml);
				break;
				}

			break;
			   
		case ParseRadio:
			if (IsEqual(name, "radio"))
				{
				widget = uigtk_radio_button_create(node, parentwidget);
				uixml_parsename(node, widget, bondxml);
				break;
				}
			break;
		case ParseBox:
			if (IsEqual(name, "cell"))
				{
				widget = uixml_parse(node, ParseGeneral, 
								parentwidget, bondxml);
				uixml_generatepath(node, bondxml, &tmpstr1, &tmpstr2);
				if (widget)
					uigtk_box_add(node, widget, parentwidget,
					              poscounter++); 
				break;
				}
			break;
		case ParseVHBox:
			if (IsEqual(name, "cell"))
				{
				widget = uixml_parse(node, ParseGeneral, parentwidget,
				                bondxml);
				uixml_generatepath(node, bondxml, &tmpstr1, &tmpstr2);
				if (widget)
					uigtk_vhbox_add(node, widget, parentwidget);
				break;
				}
			break;
		case ParseNotebook:
			if (IsEqual(name, "page"))
				{
				widget =
				    uixml_parse(node, ParseGeneral, parentwidget,
				                bondxml);
				uixml_generatepath(node, bondxml, &tmpstr1, &tmpstr2);
				if (widget)
					uigtk_notebook_addpage(node, widget, parentwidget);
				break;
				}
			break;
		case ParseMenuBar:
			if (IsEqual(name, "menu"))
				{
				widget = uigtk_menu_create(node, parentwidget, bondxml);
				uixml_parsename(node, widget, bondxml);
				if (widget)
					uigtk_menubar_add(widget, parentwidget);
				break;
				}
			break;
		default:
			/* Unknown state */
			errormsg("Unknown state %d\n", state);
			}

		
		/* add the widget to its parent if its a special type of container
		   widget */
		if (state == ParseToolBar && widget)
			{
			uigtk_toolbar_add(widget, parentwidget);
			widget = NULL;
			}
		if (state == ParseHandleBox && widget)
			{
			gtk_container_add(GTK_CONTAINER(parentwidget), widget);
			widget = NULL;
			}
		}
	/* possible memory leak if for example 2 cells occur in same table */
	if (tmpstr1)
		g_free(tmpstr1);
	if (tmpstr2)
		g_free(tmpstr2);
	if (scrolledwindow)
		widget = scrolledwindow;
	return widget;
	}

/**
 * uixml_parsename:
 * @node: Xml node
 * @widget: Widget to get name for.
 *
 * Work out the name of widgets and add to hash table.
 */
gint
uixml_parsename(xmlNode * node, UiWidget * widget, UiXML * uixml)
	{
	gchar *longname, *shortname = NULL;

	g_assert(widget);

	uixml_parsewidget_attribute(node, widget, &shortname, uixml);

	uixml_generatepath(node, uixml, &longname, &shortname);
	g_hash_table_insert(uixml->widgets, widget, longname);
	g_hash_table_insert(uixml->shortname, shortname, widget);
	g_hash_table_insert(uixml->longname, longname, widget);
	return 0;
	}

static void
uixml_buildglistonmatch(void *_key, void *_widget, void *_retlist)
	{
	UiWidget *widget;
	gchar *key;
	gchar *comparestr;
	guint i = 0;
	gboolean fail = FALSE;
	GList *retlist;

	retlist = _retlist;
	widget = _widget;
	key = _key;
	comparestr = (g_list_first(retlist))->data;

	while (comparestr[i] != 0)
		{
		if (comparestr[i] != key[i])
			{
			fail = TRUE;
			break;
			}
		i++;
		}
	if (fail == FALSE)
		retlist = g_list_append(retlist, widget);

	/* Andru, you are norty.  You don't want to return that! */
	/* return retlist; */
	}

GList *
uixml_widgetchildren(UiXML * bondxml, gchar * longname)
	{
	GList *retlist;

	/* GList *walk; */

	g_assert(longname);
	retlist = g_list_append(NULL, longname);
	g_hash_table_foreach(bondxml->longname, uixml_buildglistonmatch, retlist);
	retlist = g_list_remove_link(retlist, retlist);
	/* for(walk=g_list_first(retlist);walk!=NULL;walk=walk->next)
	   debugmsg("%s",g_hash_table_lookup(bond->widgets, walk->data)); */ 
	return retlist;
	}

gchar *
uixml_getwidgetshortname_fromname(gchar * name)
	{
	gint i, slen;

	if (!name)
		return NULL;
	slen = strlen(name);
	for (i = slen - 1; i >= 0; i++)
		if (name[i] == '/')
			return &name[i];
	return name;
	}
