/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include <glib.h>
#include <gtk/gtk.h>
#include <libxml/parser.h>
#include <libxml/tree.h>

#include "bc.h"
#include "uibox.h"
#include "uigtk.h"
#include "uigeneral.h"

/**
 * uigtk_box_create:
 * @widgetnode: an xml node in the bond XML file.
 * @parentwidget: widget that will be the parent of this.
 *
 * Creates a table/box based on properties specified in the XML
 * @node. It makes the widget visible by default.
 *
 * Attributes:
 * name:           Specifyies the name to be given to this widget.
 * columns:        Number of columns.
 * rows:           Number of rows.
 * homogeneous:    Specifies whether cells should all be same size. Can
 *                 be either "true" or "false".
 * visible:        "true" or "false".
 * column_spacing: Number pixels between each column.
 * row_spacing:    Number pixels between each row.
 * spacing:        Number pixels between each column and row. If this
 *                 attribute is used, will override column_spacing and
 *                 row_spacing.
 *
 * Example:        <box name="givenname" columns="2" spacing="3">...</box>
 *
 * Returns: Pointer to the widget created.
 */
GtkWidget *
uigtk_box_create(xmlNode * widgetnode, GtkWidget * parentwidget)
	{
	GtkWidget *widget;
	xmlChar *content = NULL;
	gboolean visible = TRUE;
	guint columns = 1, rows = 1;
	guint column_spacing = 4, row_spacing = 4;
	guint border = 0;
	gboolean homogeneous = FALSE;

	content = xmlGetProp(widgetnode, "columns");
	if (content)
		columns = atoi(content);

	content = xmlGetProp(widgetnode, "rows");
	if (content)
		rows = atoi(content);

	content = xmlGetProp(widgetnode, "homogeneous");
	if (content && (IsTrue(content)))
		homogeneous = TRUE;

	content = xmlGetProp(widgetnode, "visible");
	if (content && IsFalse(content))
		visible = FALSE;

	content = xmlGetProp(widgetnode, "column_spacing");
	if (content)
		column_spacing = atoi(content);

	content = xmlGetProp(widgetnode, "row_spacing");
	if (content)
		row_spacing = atoi(content);

	content = xmlGetProp(widgetnode, "spacing");
	if ((content) && (column_spacing = atoi(content)))
		{
		row_spacing = column_spacing;
		}

	content = xmlGetProp(widgetnode, "border");
	if (content)
		border = atoi(content);

	debuglog(50, "Creating table box %dx%d with border %d", columns, rows,
	         border);

	widget = gtk_table_new(rows, columns, homogeneous);
	gtk_table_set_row_spacings(GTK_TABLE(widget), row_spacing);
	gtk_table_set_col_spacings(GTK_TABLE(widget), column_spacing);

	gtk_container_set_border_width(GTK_CONTAINER(widget), border); 

	if (visible)
		gtk_widget_show(widget);

	return widget;
	}

/**
 * uigtk_box_add:
 * @widgetnode: an xml node in the bond XML file.
 * @parentwidget: widget that will be the parent of this.
 *
 * Adds a cell to a table/box based on properties specified in the XML
 * @node.
 *
 * Options:
 * xpos: column position in table.
 * ypos: row position in table
 * xpad:
 * ypad:
 * xspan: span of cell
 * yspan: span of cell
 * xexpand:
 * yexpand:
 * xfill:
 * yfill:
 * xshrink:
 * yshrink:
 *
 * Example:
 * <cell ypos="1">
 ...
 * </cell>
 *
 * Returns: Pointer to the widget created.
 */

gint
uigtk_box_add(xmlNode * widgetnode, GtkWidget * widget,
              GtkWidget * parentwidget, guint poscounter)
	{
	xmlChar *content = NULL;
	guint left_attach = 0, right_attach, xspan = 1;
	guint top_attach = 0, bottom_attach, yspan = 1;
	guint xpad = 0, ypad = 0;
	GtkAttachOptions xoptions, yoptions;

	/* Note setting expand to false on yaxis. Need to specify manually 
	 * for it */
	if (GTK_IS_TREE_VIEW(widget))
		{
		xoptions = GTK_FILL | GTK_EXPAND;
		yoptions = GTK_FILL | GTK_EXPAND;
		}
	else
		{
		xoptions = GTK_FILL | GTK_EXPAND;
		yoptions = GTK_FILL;
		}

	/* get packing position */
	/* uigtk_box_add_getpos(parentwidget, poscounter, &left_attach,
	   &top_attach); */

	if (!GTK_IS_TABLE(parentwidget))
		{
		errormsg("Error, not a table. %s", xmlGetNodePath(widgetnode));
		return -1;
		}

	content = xmlGetProp(widgetnode, "xpos");
	if (content)
		left_attach = atoi(content);

	content = xmlGetProp(widgetnode, "ypos");
	if (content)
		top_attach = atoi(content);

	content = xmlGetProp(widgetnode, "xspan");
	if (content)
		xspan = atoi(content);

	content = xmlGetProp(widgetnode, "yspan");
	if (content)
		yspan = atoi(content);

	content = xmlGetProp(widgetnode, "xpad");
	if (content)
		xpad = atoi(content);

	content = xmlGetProp(widgetnode, "ypad");
	if (content)
		ypad = atoi(content);

	content = xmlGetProp(widgetnode, "padding");
	if (content)
		{
		ypad = atoi(content);
		xpad = ypad;
		}

	content = xmlGetProp(widgetnode, "xexpand");
	if (content && IsFalse(content))
		xoptions &= ~GTK_EXPAND;

	content = xmlGetProp(widgetnode, "yexpand");
	if (content)
		{
		if (IsTrue(content))
			yoptions |= GTK_EXPAND;
		else if (IsFalse(content))
			yoptions &= ~GTK_EXPAND;
		}

	content = xmlGetProp(widgetnode, "expand");
	if (content && IsFalse(content))
		{
		xoptions &= ~GTK_EXPAND;
		yoptions |= GTK_EXPAND;
		}

	content = xmlGetProp(widgetnode, "xfill");
	if (content && IsFalse(content))
		xoptions &= ~GTK_FILL;

	content = xmlGetProp(widgetnode, "yfill");
	if (content && IsFalse(content))
		yoptions &= ~GTK_FILL;

	content = xmlGetProp(widgetnode, "fill");
	if (content && IsFalse(content))
		{
		xoptions &= ~GTK_FILL;
		yoptions &= ~GTK_FILL;
		}

	content = xmlGetProp(widgetnode, "xshrink");
	if (content && IsTrue(content))
		xoptions |= GTK_SHRINK;

	content = xmlGetProp(widgetnode, "yshrink");
	if (content && IsTrue(content))
		yoptions |= GTK_SHRINK;

	content = xmlGetProp(widgetnode, "shrink");
	if (content && IsTrue(content))
		{
		xoptions |= GTK_SHRINK;
		yoptions |= GTK_SHRINK;
		}

	right_attach = left_attach + xspan;
	bottom_attach = top_attach + yspan;
	if (right_attach == left_attach)
		right_attach++;
	if (top_attach == bottom_attach)
		bottom_attach++;
	debuglog(75, "Adding widget to box L%d-T%d->R%d-B%d", left_attach,
	   top_attach, right_attach, bottom_attach); 

	/* Attach the table with options set */
	gtk_table_attach(GTK_TABLE(parentwidget), widget,
	                 left_attach, right_attach,
	                 top_attach, bottom_attach, xoptions, yoptions, 
			 xpad, ypad);
	
	return 0;
	}

GtkWidget *
uigtk_vhbox_create(xmlNode * widgetnode, GtkWidget * parentwidget,
                   gint vboxtrue)
	{
	xmlChar *content = NULL;
	GtkWidget *widget;
	guint spacing = 4, border = 0;
	gboolean homogeneous = FALSE;

	content = xmlGetProp(widgetnode, "homogeneous");
	if (content)
		{
		if (IsTrue(content))
			homogeneous = TRUE;
		}

	content = xmlGetProp(widgetnode, "spacing");
	if (content)
		{
		spacing = atoi(content);
		}

	content = xmlGetProp(widgetnode, "border");
	if (content)
		{
		border = atoi(content);
		}

	if (vboxtrue == 1)
		widget = gtk_vbox_new(homogeneous, spacing);
	else
		widget = gtk_hbox_new(homogeneous, spacing);

	gtk_container_set_border_width(GTK_CONTAINER(widget), border);

	debuglog(53, "Created a box of some kind with spacing %d, border %d %p",
	         spacing, border,widget);
	if (widget->parent!=NULL)
		errormsg("Box parent is not null?? %p",widget);
	return widget;
	}

gint
uigtk_vhbox_add(xmlNode * widgetnode, GtkWidget * widget,
                GtkWidget * parentwidget)
	{
	xmlChar *content = NULL;
	guint padding = 0;
	gboolean expand = FALSE, fill = TRUE;

	/* shouldn't shrink still be there? */
	/* shrink = FALSE; */

	/* Note setting expand to false on yaxis. Need to specify manually 
	 * for it */
	if (GTK_IS_TREE_VIEW(widget) || GTK_IS_BOX(widget) 
		|| GTK_IS_TABLE(widget))
		{
		expand = TRUE;
		fill = TRUE;
		}

	content = xmlGetProp(widgetnode, "padding");
	if (content)
		{
		padding = atoi(content);
		}

	content = xmlGetProp(widgetnode, "expand");
	if (content)
		{
		if (IsFalse(content))
			expand = FALSE;
		else
			expand = TRUE;
		}

	content = xmlGetProp(widgetnode, "fill");
	if (content)
		{
		if (IsFalse(content))
			fill = FALSE;
		else
			fill = TRUE;
		}

	g_assert(parentwidget);
	g_assert(widget);
	if (widget->parent != NULL)
	     errormsg("aGr.. Cant add twice %p %p type: %d. \n",
			parentwidget,widget,GTK_WIDGET_TYPE(widget));
	   
	gtk_box_pack_start(GTK_BOX(parentwidget), widget, expand, fill, padding);
	gtk_widget_show(parentwidget);
	debuglog(50, "Added widget to hvbox with padding %d", padding);
	return 0;
	}

