
/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include <glib.h>
#include <gtk/gtk.h>
#include <libxml/parser.h>

#include "uigtk.h"
#include "xml_display.h"
#include "xml_funcs.h"
#include "bc.h"
#include "uibox.h"
#include "parse.h"
#include "uigeneral.h"
#include "uidropdown.h"

void
uigtk_addcontainer(GtkWidget * child, GtkWidget * parent)
	{
	gtk_container_add(GTK_CONTAINER(parent), child);
	}

/**
 * uigtk_window_create:
 * @widgetnode: an xml node in the bond XML file.
 * @parentwidget: widget that will be the parent of this.
 *
 * Creates a window based on properties specified in the XML
 * @node.
 *
 * Attributes:
 * name:      Specifying the name to be given to this widget.
 * title:     Title of the window.
 * visible:   Can be either "true" or "false".
 * resizable: Whether window is resizable or not.
 *
 * Example:
 * <window name="givenname" title="This is the title" visible="true">
 *
 * Returns: Pointer to the widget created.
 */
GtkWidget *
uigtk_window_create(UiXML * uixml, xmlNode * node, GtkWidget * parentwidget,
                    gboolean * visible)
	{
	gchar *content = NULL;
	gboolean resizable = TRUE;
	GtkWidget *window;

	window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	if (uixml->defaultvisible)
		*visible = TRUE;
	else
		*visible = FALSE;

	if (!(content = xmlGetProp(node, "title")))
		if (!(content = xmlGetProp(node, "name")))
			errormsg("Window contains neither a title nor name attribute");

	gtk_window_set_title(GTK_WINDOW(window), content);
	debuglog(37, "Creating window with title %s", content);

	content = xmlGetProp(node, "resizable");
	if (content && IsTrue(content))
		resizable = TRUE;
	else if (content && IsFalse(content))
		resizable = FALSE;
	gtk_window_set_resizable(GTK_WINDOW(window), resizable);

	content = xmlGetProp(node, "destroy_with_parent");
	if (content)
		{
		if (IsTrue(content))
			gtk_window_set_destroy_with_parent(GTK_WINDOW(window), TRUE);
		else if (IsFalse(content))
			gtk_window_set_destroy_with_parent(GTK_WINDOW(window), FALSE);
		}

	content = xmlGetProp(node, "visible");
	if (content && IsTrue(content))
		*visible = TRUE;
	else if (content && IsFalse(content))
		*visible = FALSE;

	return window;
	}

/**
 * uigtk_label_create:
 * @widgetnode: a xml node in the bond XML file.
 * @parentwidget: widget that will be the parent of this.
 *
 * Creates a label based on properties specified in the XML
 * @node. It makes the widget visible by default.
 *
 * Options:
 *
 *
 * Example:
 * <label>label txt</label>
 *
 * Returns: Pointer to the widget created.
 */
GtkWidget *
uigtk_label_create(xmlNode * widgetnode, GtkWidget * parentwidget)
	{
	GtkWidget *label;
	gchar *content = NULL;

	content = xmlNodeGetContent(widgetnode);
	if (content)
		label = gtk_label_new(content);
	else
		label = gtk_label_new("");

	content = xmlGetProp(widgetnode, "wrap");
	if (content && IsTrue(content))
		gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);

	/* gtk_misc_set_padding(GTK_MISC(label), 2, 0); */

	content = xmlGetProp(widgetnode, "justify");
	if (content)
		{
		if (IsEqual(content, "left"))
			{
			/* gtk_misc_set_alignment (GTK_MISC(label), 0,0); */
			gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);
			}
		else if (IsEqual(content, "right"))
			{
			/* gtk_misc_set_alignment (GTK_MISC(label), 1,0); */
			gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_RIGHT);
			}
		else if (IsEqual(content, "center"))
			gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_CENTER);
		else if (IsEqual(content, "fill"))
			gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_FILL);
		}
	else
		gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT);

	/* debuglog(50, "Creating label to say '%s'", content); */
	/* todo, check visibility */
	gtk_widget_show(label);
	return label;
	}

GtkWidget *
uigtk_scroll_bar_create(xmlNode * widgetnode, GtkWidget * parentwidget)
	{
	GtkWidget *scrollbar = NULL;
	GtkAdjustment *adjust;
	gchar *content = NULL;

	adjust = GTK_ADJUSTMENT(gtk_adjustment_new(1.1, 10.6, 5.4, 0.4, 0.5, 3.5));

	content = xmlGetProp(widgetnode, "alignment");
	if (content)
		{
		if (IsEqual(content, "horizontal"))
			scrollbar = gtk_hscrollbar_new(adjust);
		else if (IsEqual(content, "vertical"))
			scrollbar = gtk_vscrollbar_new(NULL);
		}

	if (scrollbar)
		gtk_widget_show(scrollbar);
	else
		errormsg("Error processing scrollbar");
	return scrollbar;
	}

/**
 * uigtk_scale_create:
 * @widgetnode: a xml node in the bond XML file.
 * @parentwidget: widget that will be the parent of this.
 *
 * Creates a scale widget based on properties specified in the XML
 * @node. It makes the widget visible by default.
 *
 * Options:
 * <minvalue> minimum value of the scale. set to 0 by default. 
 * <maxvalue> should be greater then 0.
 * <alignment> indicates if the scale is horizontal or vertical
 * <step> sould be between the min value and the maxvalue
 * Returns: Pointer to the widget created.
 */
GtkWidget *
uigtk_scale_create(xmlNode * widgetnode, GtkWidget * parentwidget)
	{
	GtkWidget *scale = NULL;
	gdouble min = 0;
	gdouble max = 0;
	gdouble step = 0;
	gchar *content = NULL;

	content = xmlGetProp(widgetnode, "minvalue");
	if (content)
		min = atof(content);

	content = xmlGetProp(widgetnode, "maxvalue");
	if (content)
		max = atof(content);

	content = xmlGetProp(widgetnode, "step");
	if (content)
		step = atof(content);

	content = xmlGetProp(widgetnode, "alignment");
	if (content)
		{
		if (IsEqual(content, "vertical"))
			scale = gtk_vscale_new_with_range(min, max, step);
		else
			scale = gtk_hscale_new_with_range(min, max, step);
		}

	if (scale)
		gtk_widget_show(scale);
	else
		debugmsg("failed to create scale");
	return scale;

	}

GtkWidget *
uigtk_status_bar_create(xmlNode * widgetnode, GtkWidget * parentwidget)
	{
	GtkWidget *statusbar;
	gchar *content = NULL;
	xmlNodePtr node;
	guint message_context = 0;

	statusbar = gtk_statusbar_new();
	gtk_widget_show(statusbar);

	for (node = widgetnode->children; node != NULL; node = node->next)
		{
		content = xmlNodeGetContent(node);
		if (!content)
			continue;
		if (uixml_node(node, "message"))
			{
			message_context =
			    gtk_statusbar_get_context_id(GTK_STATUSBAR(statusbar),
			                                 content);
			gtk_statusbar_push(GTK_STATUSBAR(statusbar), message_context,
			                   content);
			}
		}
	return statusbar;

	}

GtkWidget *
uigtk_entry_create(UiXML * bondxml, xmlNode * widgetnode,
                   GtkWidget * parentwidget)
	{
	guint len;
	gchar *content = NULL;
	GtkWidget *widget;

	widget = gtk_entry_new();
	gtk_widget_show(widget);

	content = xmlGetProp(widgetnode, "name");
	if (content)
		debuglog(50, "Creating entry box %s", content);
	else
		debuglog(50, "Creating entry box %s", parentwidget->name);

	/* Get attributes, this is new bond */
	content = xmlGetProp(widgetnode, "defaultvalue");
	if (content)
		gtk_entry_set_text(GTK_ENTRY(widget), content);

	content = xmlGetProp(widgetnode, "visible");
	if (content && IsFalse(content))
		gtk_widget_hide(widget);

	content = xmlGetProp(widgetnode, "hide-input");
	if (content && IsTrue(content))
		gtk_entry_set_visibility(GTK_ENTRY(widget), FALSE);

	content = xmlGetProp(widgetnode, "alignment");
	if (content && atoi(content) == 1)
		gtk_entry_set_alignment(GTK_ENTRY(widget), 1);
	   
	content = xmlGetProp(widgetnode, "maxlength");
	if (content)
		{
		len = atoi(content);
		gtk_entry_set_max_length(GTK_ENTRY(widget), len);
		}
	content = xmlGetProp(widgetnode, "activates_default");
	if (content)
		{
		if (IsTrue(content))
			gtk_entry_set_activates_default(GTK_ENTRY(widget), TRUE);
		else if (IsFalse(content))
			gtk_entry_set_activates_default(GTK_ENTRY(widget), FALSE);
		}

	content = xmlGetProp(widgetnode, "editable");
	if (content)
		{
		if (IsTrue(content))
			gtk_editable_set_editable(GTK_EDITABLE(widget), TRUE);
		else if (IsFalse(content))
			gtk_editable_set_editable(GTK_EDITABLE(widget), FALSE);
		}

	return widget;
	}

gchar *
uigtk_getname(UiXML * bondxml, UiWidget * widget)
	{
	return g_hash_table_lookup(bondxml->widgets, widget);
	}

UiWidget *
uigtk_getwidget(UiXML * bondxml, gchar * name)
	{
	UiWidget *widget;

	widget = g_hash_table_lookup(bondxml->longname, name);
	if (!widget)
		return g_hash_table_lookup(bondxml->shortname, name);
	return widget;
	}

gchar *
uigtk_getwidgetshortname(UiXML * bondxml, UiWidget * widget)
	{
	return
	    uixml_getwidgetshortname_fromname(g_hash_table_lookup
	                                      (bondxml->widgets, widget));
	}

gchar *
uigtk_widgetlongname(UiXML * bondxml, gchar * widgetname)
	{
	GtkWidget *widget;

	widget = uigtk_getwidget(bondxml, widgetname);
	if (!widget)
		return NULL;
	return g_hash_table_lookup(bondxml->widgets, widget);
	}

void
uigtk_showwidget(UiXML * bondxml, UiWidget * widget)
	{
	g_assert(widget);
	gtk_widget_show(widget);
	}

void
uigtk_hidewidget(UiXML * bondxml, UiWidget * widget)
	{
	g_assert(widget);
	gtk_widget_hide(widget);
	}

UiWidgetType
uigtk_gettype(UiXML * bondxml, UiWidget * widget)
	{
	GtkType widgettype;

	g_assert(widget);

	widgettype = GTK_WIDGET_TYPE(widget);
	if (widgettype == GTK_TYPE_ENTRY)
		return UiEntryBox;
	if (widgettype == GTK_TYPE_COMBO)
		return UiDropDown;
	if (widgettype == GTK_TYPE_TREE_VIEW)
		return UiGrid;
	if (widgettype == GTK_TYPE_TREE_VIEW_COLUMN)
		return UiGridColumn;
	if (widgettype == GTK_TYPE_SCROLLED_WINDOW)
		return UiScrolledWindow;
	if (widgettype == GTK_TYPE_CHECK_BUTTON)
		return UiCheckButton;
	if (widgettype == GTK_TYPE_SPIN_BUTTON)
		return UiSpinButton;
	if (widgettype == GTK_TYPE_LABEL)
		return UiLabel;
	if (widgettype == GTK_TYPE_TOGGLE_BUTTON)
		return UiToggleButton;
	if (widgettype == GTK_TYPE_RADIO_BUTTON)
		return UiRadioButton;
	if (widgettype == GTK_TYPE_WINDOW)
		return UiForm;
	if (widgettype == GTK_TYPE_FRAME)
		return UiFrame;
	if (widgettype == GTK_TYPE_TABLE)
		return UiFrame;
	if (widgettype == GTK_TYPE_BOX)
		return UiFrame;
	if (widgettype == GTK_TYPE_VBOX)
		return UiFrame;
	if (widgettype == GTK_TYPE_HBOX)
		return UiFrame;
	if (widgettype == GTK_TYPE_HANDLE_BOX)
		return UiGeneral;
	if (widgettype == GTK_TYPE_TOOLBAR)
		return UiGeneral;
	if (widgettype == GTK_TYPE_HANDLE_BOX)
		return UiFrame;
	if (widgettype == GTK_TYPE_NOTEBOOK)
		return UiNotebook;
	if (widgettype == GTK_TYPE_BUTTON)
		return UiButton;
	if (widgettype == GTK_TYPE_TOOL_BUTTON)
		return UiButton;
	if (widgettype == GTK_TYPE_TEXT_VIEW)
		return UiTextBox;
	if (widgettype == GTK_TYPE_TEXT_BUFFER)
		return UiTextBox;
	if (widgettype == GTK_TYPE_MENU_ITEM)
		return UiMenuItem;
	if (widgettype == GTK_TYPE_IMAGE_MENU_ITEM)
		return UiMenuItem;
	if (widgettype == GTK_TYPE_MENU_BAR)
		return UiGeneral;
	if (widgettype == GTK_TYPE_HSEPARATOR)
		return UiGeneral;
	if (widgettype == GTK_TYPE_VSEPARATOR)
		return UiGeneral;
	warningmsg("Write me!, i dont understand this type %d on widget %s",
	         widgettype, uigtk_getname(bondxml, widget));

	return UiGeneral;
	}

/* Must be a long name passed in */
GList *
uigtk_getchildren(UiXML * bondxml, gchar * widgetname)
	{
	gchar *tmpstr;
	GList *retlist;

	if (!widgetname || strlen(widgetname) == 0)
		tmpstr = (gchar *) g_strdup("");
	else
		tmpstr = (gchar *) g_strdup_printf("%s/", widgetname);

	retlist = uixml_widgetchildren(bondxml, tmpstr);
	g_free(tmpstr);

	return retlist;
	}

/**
 * uigtk_dropdown_getparent:
 *
 *
 */
UiWidget *
uigtk_getparent(UiXML * bondxml, UiWidget * widget)
	{
	gint i;
	gchar *widgetname;
	GtkWidget *retwidget = NULL;

	g_assert(widget);
	widgetname = g_strdup(uigtk_getname(bondxml, widget));

	for (i = strlen(widgetname) - 1; i >= 0; i--)
		{
		if (widgetname[i] == '/')
			{
			widgetname[i] = 0;
			retwidget = uigtk_getwidget(bondxml, widgetname);
			break;
			}
		widgetname[i] = 0;
		}
	g_free(widgetname);
	return retwidget;
	}

/**
 * uigtk_getvalue:
 * 
 * todo, someone needs to say something.
 *
 */
gint
uigtk_getvalue(UiXML * bondxml, UiWidget * widget, gchar ** value)
	{
	static gchar tval[] = "t";
	static gchar fval[] = "f";
	gboolean state;
	GtkType widgettype;
	GtkTextIter start, end;
	GtkTextBuffer *buffer;

	g_assert(widget);

	widgettype = GTK_WIDGET_TYPE(widget);
	if (widgettype == GTK_TYPE_ENTRY)
		{
		*value = (gchar *) gtk_entry_get_text(GTK_ENTRY(widget));
		}
	else if (widgettype == GTK_TYPE_COMBO)
		{
		*value =
		    (gchar *) gtk_entry_get_text(GTK_ENTRY(GTK_COMBO(widget)->entry));
		}
	else if (widgettype == GTK_TYPE_TEXT_VIEW)
		{
		buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(widget));
		/* buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(widget)); */
		gtk_text_buffer_get_iter_at_offset(GTK_TEXT_BUFFER(buffer), &start, 0);
		gtk_text_buffer_get_iter_at_offset(GTK_TEXT_BUFFER(buffer), &end, -1);
		*value =
		    (gchar *) gtk_text_buffer_get_text(GTK_TEXT_BUFFER(buffer), &start,
		                                       &end, FALSE);
		/* m.essage("Value %s got", *value); */
		}
	else if (widgettype == GTK_TYPE_SPIN_BUTTON)
		{
		GtkSpinButton *spin = GTK_SPIN_BUTTON(widget);

		*value = (gchar *) gtk_entry_get_text(&spin->entry);
		}
	else if (widgettype == GTK_TYPE_CHECK_BUTTON
	         || widgettype == GTK_TYPE_TOGGLE_BUTTON
	         || widgettype == GTK_TYPE_RADIO_BUTTON)
		{
		state = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget));
		if (state)
			*value = tval;
		else
			*value = fval;
		/* m.essage("returned %s", *value); */
		}
	else
		{
		*value = NULL;
		errormsg
		("This widget: '%s' hasn't been added yet. You are welcome to code it.",
		 uigtk_getname(bondxml, widget));
		return -1;
		}

	/* else if (widgettype == GTK_TYPE_TOGGLE_BUTTON || widgettype ==
	   GTK_TYPE_CHECK_BUTTON) { state = (gint)
	   gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(widget)); *value =
	   g_strdup_printf("%d", state); #warning "Memory leak here" return 0; }
	   else if (widgettype == GTK_TYPE_RADIO_BUTTON) { return
	   uigtk_radio_getvalue(widget, value); } */
	/*
	   else if (widgettype == GTK_TYPE_TEXT) { *value =
	   gtk_editable_get_chars(GTK_EDITABLE(widget), 0, -1); return 0; } */ 
	return 0;
	}

/**
 * uigtk_setvalue:
 * @widget:
 * @value: String value to set the widget to.
 *
 * Set the value in a widget to @value.  This works on most widget types, expect
 * for a few for obvious reasons, like clists.
 *
 */
gint
uigtk_setvalue(UiXML * bondxml, UiWidget * widget, gchar * value)
	{
	gboolean freevalue = FALSE, state;
	GtkType widgettype;
	GtkTextBuffer *buffer;

	g_assert(widget);
	if (!value)
		{
		value = mem_strdup("");
		freevalue = TRUE;
		}

	widgettype = GTK_WIDGET_TYPE(widget);
	if (widgettype == GTK_TYPE_ENTRY)
		gtk_entry_set_text(GTK_ENTRY(widget), value);

	else if (widgettype == GTK_TYPE_TEXT_VIEW)
		{
		buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(widget));
		gtk_text_buffer_set_text(GTK_TEXT_BUFFER(buffer), value, -1);
		gtk_text_view_set_buffer(GTK_TEXT_VIEW(widget),
		                         GTK_TEXT_BUFFER(buffer));
		}

	else if (widgettype == GTK_TYPE_LABEL)
		gtk_label_set_text(GTK_LABEL(widget), value);

	else if (widgettype == GTK_TYPE_SPIN_BUTTON)
		gtk_spin_button_set_value(GTK_SPIN_BUTTON(widget), atof(value));

	else if (widgettype == GTK_TYPE_CHECK_BUTTON ||
	         widgettype == GTK_TYPE_TOGGLE_BUTTON ||
	         widgettype == GTK_TYPE_RADIO_BUTTON)
		{

		if (value && strcmp(value, "t") == 0)
			state = TRUE;
		else
			state = FALSE;
		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widget),
		                             (gboolean) state);
		}
	else if (widgettype == GTK_TYPE_COMBO)
		gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(widget)->entry), value);
	else
		{
		errormsg
		("This widget hasn't been added yet. You are welcome to code it. This"
		 "is for widget %s", uigtk_getname(bondxml, widget));
		if (freevalue == TRUE)
			mem_free(value);
		return -1;
		}
	if (freevalue == TRUE)
		mem_free(value);

#if 0

	else if (widgettype == GTK_TYPE_TOGGLE_BUTTON
	         || widgettype == GTK_TYPE_CHECK_BUTTON)
		{
		state = FALSE;
		if (IsEqual(value, "t") || IsEqual(value, "1") || IsEqual(value, "T")
		        || IsEqual(value, "Y"))
			state = TRUE;

		gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(widget),
		                             (gboolean) state);
		return 0;
		}
	else if (widgettype == GTK_TYPE_RADIO_BUTTON)
		return uigtk_radio_setvalue(widget, value);

	else if (widgettype == GTK_TYPE_TEXT)
		{
		gtk_editable_delete_text(GTK_EDITABLE(widget), 0, -1);
		if (value)
			gtk_editable_insert_text(GTK_EDITABLE(widget), value, strlen(value),
			                         &pos);
		gtk_editable_set_position(GTK_EDITABLE(widget), 0);
		return 0;
		}
#endif

	return 0;
	}

/**
 * uigtk_isvisible:
 *
 * Work out if a widget is visible.
 *
 */
gboolean
uigtk_isvisible(UiXML * bondxml, UiWidget * widget)
	{
	/* g_assert(widget); */
	return GTK_WIDGET_VISIBLE(widget);
	}

/**
 * uigtk_seteditable:
 * @widget: 
 * @enable: %TRUE if you want widget editable, else %FALSE
 *
 */
void
uigtk_seteditable(UiXML * bondxml, UiWidget * widget, gboolean state)
	{
	gint widgettype;
	GtkEditable *editable;

	if (GTK_IS_WIDGET(widget))
		gtk_widget_set_sensitive(widget, state);

	widgettype = uigtk_gettype(bondxml, widget);
	switch (widgettype)
		{
	case UiDropDown:
		widget = uigtk_dropdown_getentrybox(bondxml, widget);
	case UiEntryBox:
	case UiSpinButton:
		editable = GTK_EDITABLE(widget);
		gtk_editable_set_editable(editable, state);
		break;
	case UiRadioButton:
	case UiCheckButton:
	case UiToggleButton:
	case UiTextBox:
		break;
	default:
		break;
		}
	}

/**
 * uigtk_geteditable:
 * @widget: 
 * returns wether the widegt is editable or not.
 */
gint
uigtk_geteditable(UiXML * bondxml, UiWidget * widget)
	{
	GtkEditable *editable;
	gint widgettype;

	widgettype = uigtk_gettype(bondxml, widget);

	if (widgettype == UiEntryBox)
		{
		editable = GTK_EDITABLE(widget);
		return gtk_editable_get_editable(editable);
		}

	else if (widgettype == GTK_TYPE_SPIN_BUTTON)
		{
		editable = GTK_EDITABLE(widget);
		return gtk_editable_get_editable(editable);
		}

	return 1;
	}

/**
 * uigtk_setuserdata(UiWidget *item, gpointer data):
 * @widget: 
 *
 */
void
uigtk_setuserdata(UiXML * bondxml, UiWidget * item, gpointer data)
	{
	g_assert(item);

	return gtk_object_set_user_data(GTK_OBJECT(item), data);
	}

/**
 * uigtk_getcontainerchildren:
 */

GList *
uigtk_getcontainerchildren(UiXML * bondxml, UiWidget * list)
	{
	g_assert(list);

	return gtk_container_children(GTK_CONTAINER(list));
	}

GtkWidget *
uigtk_handlebox_create(UiXML * bondxml, xmlNode * widgetnode,
                       GtkWidget * parentwidget)
	{
	GtkWidget *handlebox;

	handlebox = gtk_handle_box_new();
	gtk_widget_show(handlebox);
	return handlebox;
	}

gint
uigtk_setwidgetsize(UiXML * uixml, UiWidget * widget, gint width, gint height)
	{
	if (width < -1)
		return -1;
	if (height < -1)
		return -2;
	gtk_widget_set_size_request(widget, width, height);
	return 0;
	}

gint
uigtk_raisewindow(UiXML *uixml, UiWidget *widget)
	{
   	gdk_window_raise(GTK_WIDGET(widget)->window);
	return 0;
	}

