
/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include <glib.h>
#include <gtk/gtk.h>
#include <libxml/parser.h>

#include "bc.h"
#include "uilist.h"
#include "parse.h"
#include "xml_funcs.h"

static GtkTreeViewColumn *
uihtml_list_createsecretcolumn()
	{
	struct UiColumn uic;
	GtkTreeViewColumn *column;
	GtkCellRenderer *renderer;
	gboolean visible = TRUE;

	renderer = gtk_cell_renderer_text_new();
	visible = FALSE;
	uic.sortable = FALSE;
	uic.clickable = FALSE;
	uic.resizable = FALSE;
	uic.reorderable = FALSE;
	uic.spacing = 0;
	uic.sizing = GTK_TREE_VIEW_COLUMN_FIXED;
	column =
	    gtk_tree_view_column_new_with_attributes("", renderer, "text", 0, NULL);

	gtk_tree_view_column_set_visible(column, visible);
	gtk_tree_view_column_set_clickable(column, uic.clickable);
	gtk_tree_view_column_set_resizable(column, uic.resizable);
	gtk_tree_view_column_set_reorderable(column, uic.reorderable);
	gtk_tree_view_column_set_spacing(column, uic.spacing);
	gtk_tree_view_column_set_sizing(column, uic.sizing);
	return column;
	}

/*G_TYPE_CHAR,
  G_TYPE_UCHAR,
  G_TYPE_BOOLEAN,
  G_TYPE_INT,
  G_TYPE_UINT,
  G_TYPE_LONG,
  G_TYPE_ULONG,
  G_TYPE_ENUM,
  G_TYPE_FLAGS,
  G_TYPE_FLOAT,
  G_TYPE_DOUBLE,
  G_TYPE_STRING,
  G_TYPE_POINTER,
  G_TYPE_BOXED,
  G_TYPE_PARAM,
  G_TYPE_OBJECT */
static GType
uihtml_list_getcolumn_gtype(xmlNode * widgetnode)
	{
	gchar *content;
	GType retval = G_TYPE_STRING;

	content = xmlGetProp(widgetnode, "type");
	if (content)
		{
		if (IsEqual(content, "char") || IsEqual(content, "g_type_char"))
			retval = G_TYPE_CHAR;
		else if (IsEqual(content, "boolean") || IsEqual(content, "bool")
		         || IsEqual(content, "g_type_boolean"))
			retval = G_TYPE_BOOLEAN;
		else if (IsEqual(content, "int") || IsEqual(content, "number")
		         || IsEqual(content, "g_type_int"))
			retval = G_TYPE_INT;
		else if (IsEqual(content, "uint") || IsEqual(content, "g_type_uint"))
			retval = G_TYPE_UINT;
		else if (IsEqual(content, "long") || IsEqual(content, "g_type_long"))
			retval = G_TYPE_LONG;
		else if (IsEqual(content, "ulong") || IsEqual(content, "g_type_ulong"))
			retval = G_TYPE_ULONG;
		else if (IsEqual(content, "float") || IsEqual(content, "g_type_float"))
			retval = G_TYPE_FLOAT;
		else if (IsEqual(content, "double")
		         || IsEqual(content, "g_type_double"))
			retval = G_TYPE_DOUBLE;
		else if (IsEqual(content, "object")
		         || IsEqual(content, "g_type_object"))
			retval = G_TYPE_OBJECT;
		else if (IsEqual(content, "pointer")
		         || IsEqual(content, "g_type_pointer"))
			retval = G_TYPE_POINTER;
		}
	return retval;
	}

/**
 * uihtml_list_createcolumn:
 * @widgetnode: a xml node in the bond XML file.
 * @colnumber: Reference number for the column.
 * @coltype: What type of data is stored in column.
 * @sortable: Specifiy if this is sortable.
 * 
 * This is a internal function, used by uihtml_list_create()
 * to generate the columns in a list.
 *
 * Options:
 * <title> Name of header that appears.
 * <visible> If column is hidden or not.
 * <sortable> specify if the column can be sortable.
 * <clickable> If the column can be clicked on or not.
 * <resizable> If the column is resiable or not.
 * <reorderable> If columns can be moved around or not.
 * <spacing> Pixels between columns.
 * <sizing> Can be growonly, autoresize, or fixed.
 *
 * Example:
 * <label>label txt</label>
 *
 * Returns: Pointer to the widget created.
 */
GtkTreeViewColumn *
uihtml_list_createcolumn(xmlNode * widgetnode, struct UiColumn uic,
                        UiXML * bondxml)
	{
	GtkTreeViewColumn *column = NULL;
	GtkCellRenderer *renderer;
	gboolean visible = TRUE;
	gchar *content, *title = NULL;
	xmlNodePtr parent = NULL;

	/* Get the title property from the column element */
	content = xmlGetProp(widgetnode, "title");
	if (content)
		title = content;

	/* Get the parent of the column */
	parent = widgetnode->parent;
	content = xmlGetProp(parent, "sortable");
	if (content && IsEqual(content, "true"))
		uic.sortable = TRUE;
	else
		uic.sortable = FALSE;

	content = xmlGetProp(parent, "visible");
	if (content && IsFalse(content))
		visible = FALSE;

	content = xmlGetProp(parent, "clickable");
	if (content && IsFalse(content))
		uic.clickable = FALSE;
	else
		uic.clickable = TRUE;

	content = xmlGetProp(parent, "resizable");
	if (content && IsFalse(content))
		uic.resizable = FALSE;
	else
		uic.resizable = TRUE;

	content = xmlGetProp(parent, "reorderable");
	if (content && IsEqual(content, "true"))
		uic.reorderable = TRUE;
	else
		uic.reorderable = FALSE;

	content = xmlGetProp(parent, "spacing");
	if (content)
		uic.spacing = atoi(content);

	content = xmlGetProp(parent, "sizing");
	if (content)
		{
		if (IsEqual(content, "growonly"))
			uic.sizing = GTK_TREE_VIEW_COLUMN_GROW_ONLY;
		else if (IsEqual(content, "autosize"))
			uic.sizing = GTK_TREE_VIEW_COLUMN_AUTOSIZE;
		else if (IsEqual(content, "fixed"))
			uic.sizing = GTK_TREE_VIEW_COLUMN_FIXED;
		}

	content = xmlGetProp(parent, "fixedwidth");
	if (content)
		uic.maxwidth = atoi(content);

	content = xmlGetProp(parent, "maxwidth");
	if (content)
		uic.maxwidth = atoi(content);

	content = xmlGetProp(parent, "minwidth");
	if (content)
		uic.minwidth = atoi(content);

	renderer = gtk_cell_renderer_text_new();
	/* Needs to be coded, for handling stuff other than g_type_string. */
	if (uic.coltype == G_TYPE_STRING)
		if (title)
			column =
			    gtk_tree_view_column_new_with_attributes(title, renderer, "text",
			            uic.colnumber, NULL);
		else
			column =
			    gtk_tree_view_column_new_with_attributes("", renderer, "text",
			            uic.colnumber, NULL);
	else
		{
		warningmsg("Unsupported column type.");
		return NULL;
		}

	/* set the attributes */
	if (uic.fixedwidth != -1)
		gtk_tree_view_column_set_fixed_width(column, uic.fixedwidth);
	if (uic.minwidth != -1)
		gtk_tree_view_column_set_min_width(column, uic.minwidth);
	if (uic.maxwidth != -1)
		gtk_tree_view_column_set_max_width(column, uic.maxwidth);
	gtk_tree_view_column_set_visible(column, visible);
	gtk_tree_view_column_set_clickable(column, uic.clickable);
	if (uic.clickable == TRUE)
		gtk_tree_view_column_clicked(column);
	gtk_tree_view_column_set_resizable(column, uic.resizable);
	gtk_tree_view_column_set_reorderable(column, uic.reorderable);
	if (uic.sortable == TRUE)
		{
		gtk_tree_view_column_set_sort_indicator(column, TRUE);
		gtk_tree_view_column_set_sort_column_id(column, uic.colnumber);
		}
	gtk_tree_view_column_set_spacing(column, uic.spacing);
	gtk_tree_view_column_set_sizing(column, uic.sizing);
	/* setup name for widget, so bondxml sees it */

	/* if (title != NULL) xmlNewProp(widgetnode, "name", title); */
	/* Warning, because a column isn't a widget it can not be properly extracted
	   from the hash tables and treated as a widget. */

	uixml_parsename(widgetnode, (UiHtml *) column, bondxml);
	/* done */

	return column;
	}

static void
uihtml_colfree(void *ptr)
	{
	mem_free(ptr);
	}

/**
 * uihtml_list_create:
 * @widgetnode: a xml node in the bond XML file.
 * @parentwidget: widget that will be the parent of this.
 *
 * Creates a label based on properties specified in the XML
 * @node. It makes the widget visible by default.  
 *
 * Options:
 * <source>
 * <column> See details in uihtml_list_createcolumn() on options here.
 * <sortable> specify if the columns can be sortable.
 * <clickable> If the columns can be clicked on or not.
 * <resizable> If the columns is resiable or not
 * <reorderable> If columns can be moved around or not 
 * <spacing> Spacing in pixels between columns
 * <sizing> Can be growonly, autoresize, or fixed.
 * 
 * Example:
 * <label>label txt</label>
 *
 * Returns: Pointer to the widget created.
 */
UiHtml *
uihtml_list_create(xmlNode * widgetnode, UiHtml * parentwidget,
				   UiXML * bondxml, gboolean * visible)
	{
	UiHtml *list;
	
	guint colcount = 1, spacing;
	xmlNode *node;
	gchar *content = NULL;
	
	debuglog(50, "Creating a list %s", widgetnode->name);
	
	list = uihtml_create(widgetnode);
	list->widgettype = UiGeneral;

	list->openinghtml = mem_strdup("<table class=\"list\"");
	
	content = xmlGetProp(widgetnode, "sortable");
	if (content && IsTrue(content))
	  ;
	else
	  ;

	content = xmlGetProp(widgetnode, "visible");
	if (content && IsFalse(content))
	  *visible = FALSE;
	else 
	  *visible = TRUE;
	
	content = xmlGetProp(widgetnode, "clickable");
	if (content && IsFalse(content))
	  ;
	else
	  ;

	content = xmlGetProp(widgetnode, "resizable");
	if (content && IsFalse(content))
	  ;
	else
	  ;

	content = xmlGetProp(widgetnode, "reorderable");
	if (content && IsTrue(content))
	  ;
	else
	  ;

	content = xmlGetProp(widgetnode, "spacing");
	if (content && (spacing=atoi(content)))
	  list->openinghtml = mem_strdup_printf("%s cellspacing=\"%d\"", 
											list->openinghtml, spacing);

	list->openinghtml = mem_strdup_printf("%s >", list->openinghtml);
	list->closinghtml = mem_strdup("</table><!-- list -->");

	return list;
	}

gint
uihtml_list_setcellvalue(UiXML * uixml, UiHtml * widget, gint col,
                        gulong secretid, gchar * value)
	{
	gboolean found = FALSE;
	gulong gotvalue = 0;
	GValue *a;
	GValue *b;

	return 0;
	}

/* Please ensure GList length is less than num columns */
void
uihtml_list_append(UiXML * uixml, UiHtml * widget, GList * values,
                  gulong secretid, gboolean append)
	{
	GList *walk;
	GtkTreeIter iter;
	GtkListStore *store;
	GValue *a;
	gint i = 1;

	/* Initial work to build widget information */
	g_assert(widget);
/*	store = GTK_LIST_STORE(uihtml_list_getstore(widget));*/
/*	g_assert(store);*/
	/* if append is true then we add the row onto the end of the list */
	if (append == FALSE)
		gtk_list_store_prepend(store, &iter);
	else
		gtk_list_store_append(store, &iter);
	/* add secret column */
	a = g_malloc0(sizeof(GValue));
	g_value_init(a, G_TYPE_ULONG);
	g_value_set_ulong(a, secretid);
	gtk_list_store_set_value(store, &iter, 0, a);
	g_value_unset(a);
	g_free(a);
	/* add rest of columns */
	for (walk = g_list_first(values); walk; walk = walk->next)
		{
		a = g_malloc0(sizeof(GValue));
		g_value_init(a, G_TYPE_STRING);
		g_value_set_string(a, walk->data);

		gtk_list_store_set_value(store, &iter, i, a);
		g_value_unset(a);
		g_free(a);
		i++;
		}
	}

static gint
uihtml_list_numcol(UiHtml * widget)
	{
	return
	  1;
	}

gboolean
uihtml_list_setselection(UiXML * uixml, UiHtml * widget, gulong secretid)
	{
	guint secret;
	GtkTreeIter iter;
	GtkTreePath *path;
	GtkTreeSelection *selection;
	GtkTreeModel *model;

	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(widget));
	/*	model = uihtml_list_getstore(widget);*/
	if (gtk_tree_model_get_iter_first(model, &iter))
		while (1)
			{
			gtk_tree_model_get(model, &iter, 0, &secret, -1);
			if (secret == secretid)
				{
				path = gtk_tree_model_get_path(model, &iter);
				gtk_tree_selection_select_path(selection, path);
				gtk_tree_path_free(path);
				return TRUE;
				}
			if (!gtk_tree_model_iter_next(model, &iter))
				break;
			}
	return FALSE;
	}

gulong
uihtml_list_getselection(UiXML * uixml, UiHtml * widget, gboolean getvalues,
                        GList ** values)
	{
	gint numcol, i, a;
	gulong retsecret = -1;
	GList *retlist = NULL;
	gchar *tmpstr;
	GtkTreeSelection *selection;
	GtkTreeModel *model;
	GtkTreeIter iter;

	selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(widget));
	if (gtk_tree_selection_get_selected(selection, &model, &iter) == TRUE)
		{
		gtk_tree_model_get(model, &iter, 0, &a, -1);
		retsecret = a;
		if (getvalues == TRUE)
			{
			*values = NULL;
			numcol = uihtml_list_numcol(widget);
			for (i = 1; i < numcol; i++)
				{
				gtk_tree_model_get(model, &iter, i, &tmpstr, -1);
				retlist = g_list_append(retlist, g_strdup(tmpstr));
				}
			*values = retlist;
			}
		}
	else
		return -2;
	return retsecret;
	}
